import uuid
from datetime import datetime
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import config
import los_files
import mox_helpers.payloads as mox_payloads
import pydantic
import uuids
from mox_helpers.mox_helper import create_mox_helper
from pydantic import Field

from os2mo_data_import.mox_data_types import Itsystem


class StamCSV(pydantic.BaseModel):
    @staticmethod
    def get_filename() -> str:
        raise NotImplementedError("must be implemented by subclass")

    @staticmethod
    def get_facet_bvn() -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def bvn(self) -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def title(self) -> str:
        raise NotImplementedError("must be implemented by subclass")

    @property
    def class_uuid(self) -> uuid.UUID:
        raise NotImplementedError("must be implemented by subclass")


class Engagementstype(StamCSV):
    engagement_type_uuid: uuid.UUID = Field(alias="EngagementstypeUUID")
    engagement_type: str = Field(alias="Engagementstype")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Engagementstype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_type"

    @property
    def bvn(self) -> str:
        return self.engagement_type

    @property
    def title(self) -> str:
        return self.engagement_type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.engagement_type_uuid


class Enhedstype(StamCSV):
    org_unit_type_uuid: uuid.UUID = Field(alias="EnhedstypeUUID")
    org_unit_type: str = Field(alias="Enhedstype")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Enhedstype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "org_unit_type"

    @property
    def bvn(self) -> str:
        return self.org_unit_type

    @property
    def title(self) -> str:
        return self.org_unit_type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.org_unit_type_uuid


class ITSystem(StamCSV):
    it_system_uuid: uuid.UUID = Field(alias="ITSystemUUID")
    name: str = Field(alias="Name")
    user_key: str = Field(alias="Userkey")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_ITSystem.csv"


class Stillingsbetegnelse(StamCSV):
    job_function_uuid: uuid.UUID = Field(alias="StillingBetUUID")
    job_function: str = Field(alias="Stillingsbetegnelse")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Stillingsbetegnelse.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "engagement_job_function"

    @property
    def bvn(self) -> str:
        return self.job_function

    @property
    def title(self) -> str:
        return self.job_function

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.job_function_uuid


class Lederansvar(StamCSV):
    responsibility_uuid: uuid.UUID = Field(alias="LederansvarUUID")
    responsibility: str = Field(alias="Lederansvar")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Lederansvar.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "responsibility"

    @property
    def bvn(self) -> str:
        return self.responsibility

    @property
    def title(self) -> str:
        return self.responsibility

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.responsibility_uuid


class Lederniveau(StamCSV):
    level_uuid: uuid.UUID = Field(alias="LederniveauUUID")
    level: str = Field(alias="Lederniveau")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Lederniveau.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "manager_level"

    @property
    def bvn(self) -> str:
        return self.level

    @property
    def title(self) -> str:
        return self.level

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.level_uuid


class Ledertype(StamCSV):
    type_uuid: uuid.UUID = Field(alias="LedertypeUUID")
    type: str = Field(alias="Ledertype")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Ledertype.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "manager_type"

    @property
    def bvn(self) -> str:
        return self.type

    @property
    def title(self) -> str:
        return self.type

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.type_uuid


class Tilknytningsrolle(StamCSV):
    association_type_uuid: uuid.UUID = Field(alias="TilknytningsrolleUUID")
    role: str = Field(alias="Tilknytningsrolle")
    loadtime: datetime = Field(alias="Loadtime")

    @staticmethod
    def get_filename() -> str:
        return "STAM_UUID_Tilknytningsrolle.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "association_type"

    @property
    def bvn(self) -> str:
        return self.role

    @property
    def title(self) -> str:
        return self.role

    @property
    def class_uuid(self) -> uuid.UUID:
        return self.association_type_uuid


StamCSVType = Union[
    Type[Engagementstype],
    Type[Enhedstype],
    Type[ITSystem],
    Type[Stillingsbetegnelse],
    Type[Lederansvar],
    Type[Lederniveau],
    Type[Ledertype],
    Type[Tilknytningsrolle],
]


class StamImporter:
    @classmethod
    async def handle_engagementstype(cls, last_import: datetime):
        """
        Process the external 'engagement type' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(Engagementstype, last_import)
        return await cls._create_classes_from_csv(Engagementstype, rows)

    @classmethod
    async def handle_enhedstype(cls, last_import: datetime):
        """
        Process the external 'unit type' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(Enhedstype, last_import)
        return await cls._create_classes_from_csv(Enhedstype, rows)

    @classmethod
    async def handle_itsystem(cls, last_import: datetime):
        """
        Process the external 'it system' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(ITSystem, last_import)
        if rows is None:
            return

        settings = config.get_config()
        mox_helper = await create_mox_helper(settings.mox_base)
        for row in rows:
            it_system = Itsystem(
                system_name=row.name,  # type: ignore
                user_key=row.user_key,  # type: ignore
            )
            it_system.organisation_uuid = str(uuids.ORG_UUID)
            json = it_system.build()
            await mox_helper.insert_organisation_itsystem(
                json,
                str(row.it_system_uuid),  # type: ignore
            )

    @classmethod
    async def handle_stillingsbetegnelse(cls, last_import: datetime):
        """
        Process the external 'job_function' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(Stillingsbetegnelse, last_import)
        return await cls._create_classes_from_csv(Stillingsbetegnelse, rows)

    @classmethod
    async def handle_lederansvar(cls, last_import: datetime):
        """
        Process the external 'Lederansvar' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(Lederansvar, last_import)
        return await cls._create_classes_from_csv(Lederansvar, rows)

    @classmethod
    async def handle_lederniveau(cls, last_import: datetime):
        """
        Process the external 'Lederniveau' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(Lederniveau, last_import)
        return await cls._create_classes_from_csv(Lederniveau, rows)

    @classmethod
    async def handle_ledertype(cls, last_import: datetime):
        """
        Process the external 'Ledertype' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(Ledertype, last_import)
        return await cls._create_classes_from_csv(Ledertype, rows)

    @classmethod
    async def handle_tilknytningsrolle(cls, last_import: datetime):
        """
        Process the external 'Tilknytningsrolle' file and create objects if file is
        newer than last import.
        """
        rows = cls._load_csv_if_newer(Tilknytningsrolle, last_import)
        return await cls._create_classes_from_csv(Tilknytningsrolle, rows)

    @classmethod
    def _load_csv_if_newer(
        cls, csv_class: StamCSVType, last_import: datetime
    ) -> Union[List[StamCSVType], None]:
        filename = csv_class.get_filename()
        fileset = los_files.get_fileset_implementation()
        try:
            modified_datetime = fileset.get_modified_datetime(filename)
        except ValueError:
            # Raised by `fileset.get_modified_datetime` if file could not be
            # found.
            return None
        else:
            if modified_datetime <= last_import:
                return None
            return los_files.read_csv(filename, csv_class)

    @classmethod
    async def _create_classes_from_csv(
        cls, csv_class: StamCSVType, rows: Optional[List[StamCSVType]] = None
    ) -> None:
        if rows is None:
            return
        settings = config.get_config()
        mox_helper = await create_mox_helper(settings.mox_base)
        facet_bvn = csv_class.get_facet_bvn()
        facet_uuid = await mox_helper.read_element_klassifikation_facet(bvn=facet_bvn)
        for row in rows:
            klasse = mox_payloads.lora_klasse(
                bvn=row.bvn,
                title=row.title,
                facet_uuid=facet_uuid,
                org_uuid=str(uuids.ORG_UUID),
                scope="TEXT",
            )
            await mox_helper.insert_klassifikation_klasse(klasse, str(row.class_uuid))

    async def run(self, last_import: datetime):
        """
        Import all klasser and itsystems from files
        Imports based on predetermined UUIDs, so these functions should be idempotent

        Only imports if file has been modified after last_import
        """
        print("Starting STAM import")
        await self.handle_engagementstype(last_import)
        await self.handle_enhedstype(last_import)
        await self.handle_itsystem(last_import)
        await self.handle_stillingsbetegnelse(last_import)
        await self.handle_lederansvar(last_import)
        await self.handle_lederniveau(last_import)
        await self.handle_ledertype(last_import)
        await self.handle_tilknytningsrolle(last_import)
        print("STAM import done")



import click
from fix_departments import unit_fixer
from sd_changed_at import changed_at
from sd_importer import cli as sd_importer
from sync_job_id import sync_jobid
from test_mo_against_sd import cli as mo_against_sd
from test_sd_connectivity import check_connectivity

from tools.data_fixers.sd_fixup import cli as sd_fixup


@click.group()
def SDTool():
    """Common entrypoint to SD programs."""
    pass


SDTool.add_command(sd_importer, "sd_importer")
SDTool.add_command(mo_against_sd, "test_mo_against_sd")
SDTool.add_command(check_connectivity)
SDTool.add_command(unit_fixer)
SDTool.add_command(sync_jobid)
SDTool.add_command(changed_at)
SDTool.add_command(sd_fixup)


if __name__ == "__main__":
    SDTool()



from datetime import datetime

import click
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_setting
from tqdm import tqdm

from integrations.ad_integration.ad_common import AD
from integrations.ad_integration.ad_reader import ADParameterReader


class CompareEndDate(ADParameterReader):
    def __init__(self, enddate_field, uuid_field, settings=None):
        super().__init__(all_settings=settings)
        self.helper = MoraHelper(
            hostname=self.all_settings["global"]["mora.base"], use_cache=False
        )
        self.enddate_field = enddate_field
        self.uuid_field = uuid_field
        self.cpr_field = self.all_settings["primary"]["cpr_field"]

    def scan_ad(self):
        # Find all AD users whose end date is in the year 9999.
        # This does not in itself indicate an error, as users with current or
        # future engagements without an end date will have an AD end date of
        # 9999-12-31.
        bp = self._ps_boiler_plate()
        cmd = (
            self._build_user_credential()
            + f"Get-ADUser -Filter '{self.enddate_field} -like \"9999-*\"'"
            + bp["complete"]
            + f" -Properties cn,{self.uuid_field},{self.enddate_field}"
            + " | ConvertTo-Json"
        )
        return map(
            lambda ad_user: (
                ad_user["CN"],
                ad_user[self.enddate_field],
                ad_user.get(self.uuid_field),
            ),
            self._run_ps_script(cmd),
        )

    def get_mo_engagements(self, mo_user_uuid):
        return [
            {
                "from": eng["validity"]["from"],
                "to": eng["validity"]["to"],
                "is_primary": eng["is_primary"],
            }
            for eng in self.helper.read_user_engagement(
                mo_user_uuid,
                calculate_primary=True,
                read_all=True,
                skip_past=False,
            )
        ]

    def compare_mo(self):
        # Compare AD users to MO users
        print("Find users from AD")
        ad_hits = list(self.scan_ad())

        print("Check MO engagements")
        users = {}
        for cn, end_date, mo_user_uuid in tqdm(ad_hits, unit="user"):
            if mo_user_uuid:
                mo_engagements = self.get_mo_engagements(mo_user_uuid)
                if mo_engagements:
                    if not any([eng["to"] is None for eng in mo_engagements]):
                        users[mo_user_uuid] = {
                            "cn": cn,
                            "mo_engagements": mo_engagements,
                        }
        return users


class UpdateEndDate(AD):
    def __init__(self, enddate_field, uuid_field, cpr_field, settings=None):
        super().__init__(all_settings=settings)
        self.enddate_field = enddate_field
        self.uuid_field = uuid_field
        self.cpr_field = cpr_field

    def get_correct_end_date(self, doc):
        candidates = sorted(
            [val["to"] for val in doc if val["is_primary"]],
            key=lambda to_date: datetime.strptime(to_date, "%Y-%m-%d").date(),
            reverse=True,
        )
        return candidates[0]

    def get_changes(self, users):
        for uuid, doc in users.items():
            end_date = self.get_correct_end_date(doc["mo_engagements"])
            yield uuid, end_date

    def get_update_cmd(self, uuid, end_date):
        cmd_f = """
        Get-ADUser %(complete)s -Filter '%(uuid_field)s -eq "%(uuid)s"' |
        Set-ADUser %(credentials)s -Replace @{%(enddate_field)s="%(end_date)s"} |
        ConvertTo-Json
        """
        cmd = cmd_f % dict(
            uuid=uuid,
            end_date=end_date,
            enddate_field=self.enddate_field,
            uuid_field=self.uuid_field,
            complete=self._ps_boiler_plate()["complete"],
            credentials=self._ps_boiler_plate()["credentials"],
        )
        return cmd

    def run(self, cmd):
        return self._run_ps_script("%s\n%s" % (self._build_user_credential(), cmd))


@click.command()
@click.option(
    "--enddate-field",
    default=load_setting("integrations.ad_writer.fixup_enddate_field"),
)
@click.option("--uuid-field", default=load_setting("integrations.ad.write.uuid_field"))
@click.option("--dry-run", is_flag=True)
def cli(enddate_field, uuid_field, dry_run):
    """Fix enddates of terminated users.
    AD-writer does not support writing enddate of a terminated employee,
    this script finds and corrects the enddate in AD of terminated engagements.
    """

    c = CompareEndDate(enddate_field, uuid_field)
    users = c.compare_mo()
    u = UpdateEndDate(enddate_field, uuid_field, c.cpr_field)
    users = u.get_changes(users)

    for uuid, end_date in tqdm(users, unit="user", desc="Changing enddate in AD"):
        cmd = u.get_update_cmd(uuid, end_date)
        if dry_run:
            print("Command to run: ")
            print(cmd)
        else:
            result = u.run(cmd)
            if result:
                print("Result: %r" % result)

    print("All done")


if __name__ == "__main__":
    cli()



import unittest
from unittest.mock import patch
from uuid import uuid4

from helpers import dummy_settings
from parameterized import parameterized

from integrations.os2sync.os2mo import get_work_address
from integrations.os2sync.os2mo import is_ignored
from integrations.os2sync.os2mo import kle_to_orgunit
from integrations.os2sync.os2mo import partition_kle


class MockOs2moGet:
    """Class which allows patching to have a json() method"""

    def __init__(self, return_value):
        self.return_value = return_value

    def json(self):
        return self.return_value


class TestsMOAd(unittest.TestCase):
    def test_is_ignored(self):
        settings = dummy_settings
        il1, il2, iu1, iu2 = [uuid4() for u in range(4)]
        settings.os2sync_ignored_unit_levels = [il1, il2]
        settings.os2sync_ignored_unit_types = [iu1, iu2]
        unit = {
            "org_unit_level": {"uuid": str(uuid4())},
            "org_unit_type": {"uuid": str(uuid4())},
        }
        self.assertFalse(is_ignored(unit, settings))
        unit = {
            "org_unit_level": {"uuid": str(uuid4())},
            "org_unit_type": {"uuid": str(iu2)},
        }
        self.assertTrue(is_ignored(unit, settings))
        unit = {
            "org_unit_level": {"uuid": str(il1)},
            "org_unit_type": {"uuid": str(uuid4())},
        }
        self.assertTrue(is_ignored(unit, settings))

    @parameterized.expand(
        [
            ({"os2mo_has_kle": True}, ["2", "4", "6"], []),
            (
                {"os2mo_has_kle": True, "os2sync_use_contact_for_tasks": True},
                ["2", "4"],
                ["6"],
            ),
        ]
    )
    def test_kle_to_orgunit(
        self, testsettings, expected_tasks, expected_contactfortasks
    ):

        kles = [
            {
                "uuid": 1,
                "kle_aspect": [{"name": "Udførende"}],
                "kle_number": {"uuid": "2"},
            },
            {
                "uuid": 3,
                "kle_aspect": [{"name": "Udførende"}],
                "kle_number": {"uuid": "4"},
            },
            {
                "uuid": 5,
                "kle_aspect": [{"name": "Ansvarlig"}],
                "kle_number": {"uuid": "6"},
            },
        ]
        settings = dummy_settings
        settings.os2sync_use_contact_for_tasks = (
            True if testsettings.get("os2sync_use_contact_for_tasks") else False
        )

        tasks, contact_for_tasks = partition_kle(
            kles, settings.os2sync_use_contact_for_tasks
        )
        self.assertListEqual(expected_tasks, tasks)
        self.assertListEqual(expected_contactfortasks, contact_for_tasks)
        org_unit = {}
        kle_to_orgunit(
            org_unit, kles, use_contact_for_tasks=settings.os2sync_use_contact_for_tasks
        )
        self.assertListEqual(expected_tasks, org_unit.get("Tasks"))
        self.assertListEqual(
            expected_contactfortasks, org_unit.get("ContactForTasks", [])
        )

    @parameterized.expand(
        [
            (["Henvendelsessted", "Adresse"], "Henvendelsesstednavn"),
            (["Adresse", "Henvendelsessted"], "Adressenavn"),
            ([], None),
        ]
    )
    def test_get_work_address(self, work_address_names, expected):
        addresses = [
            {
                "name": "Henvendelsesstednavn",
                "address_type": {"name": "Henvendelsessted"},
            },
            {"name": "Adressenavn", "address_type": {"name": "Adresse"}},
            {"name": "Adressenavn2", "address_type": {"name": "Adresse"}},
        ]
        positions = [{"is_primary": True, "OrgUnitUuid": "Some_unit_uuid"}]
        with patch(
            "integrations.os2sync.os2mo.os2mo_get", return_value=MockOs2moGet(addresses)
        ):

            work_address = get_work_address(positions, work_address_names)
            self.assertEqual(work_address, expected)



#
# Copyright (c) 2018, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from uuid import UUID
from operator import itemgetter
import requests
from more_itertools import first
from more_itertools import one


from constants import AD_it_system
from exporters.utils.priority_by_class import choose_public_address
from integrations.os2sync.config import get_os2sync_settings
from integrations.os2sync import config
from integrations.os2sync.templates import Person
from integrations.os2sync.templates import User
from ra_utils.headers import TokenSettings

logger = logging.getLogger(config.loggername)


@lru_cache
def get_mo_session():
    session = requests.Session()
    session.verify = get_os2sync_settings().os2sync_ca_verify_os2mo
    session.headers = {
        "User-Agent": "os2mo-data-import-and-export",
    }
    session_headers = TokenSettings().get_headers()
    if session_headers:
        session.headers.update(session_headers)
    return session




# Actually recursive, but couple of levels here
JSON_TYPE = Dict[str, Union[None, str, Dict[str, Union[None, str, Dict[str, Any]]]]]


class IT:
    def __init__(self, system_name: str, user_key: str):
        self.system_name = system_name
        self.user_key = user_key

    @classmethod
    def from_mo_json(cls, response: List[JSON_TYPE]):
        """
        Designed to parse the response from MO when requesting a users it-systems
        Concretely, expects a list of this style of object:

        {'itsystem': {'name': 'Active Directory',
              'reference': None,
              'system_type': None,
              'user_key': 'Active Directory',
              'uuid': 'a1608e69-c422-404f-a6cc-b873c50af111',
              'validity': {'from': '1900-01-01', 'to': None}},
         'org_unit': None,
         'person': {'givenname': 'Solveig',
                    'name': 'Solveig Kuhlenhenke',
                    'nickname': '',
                    'nickname_givenname': '',
                    'nickname_surname': '',
                    'surname': 'Kuhlenhenke',
                    'uuid': '23d2dfc7-6ceb-47cf-97ed-db6beadcb09b'},
         'user_key': 'SolveigK',
         'uuid': 'a2fb2581-c57a-46ad-8a21-30118a3859b7',
         'validity': {'from': '2003-08-13', 'to': None}}
        """
        return [
            cls(
                system_name=it_obj["itsystem"]["name"].strip(),
                user_key=it_obj["user_key"].strip(),
            )
            for it_obj in response
        ]

    def __repr__(self):
        return "{}(system_name={},user_key={})".format(
            self.__class__.__name__, self.system_name, self.user_key
        )

    def __eq__(self, other):
        if not isinstance(other, IT):
            raise TypeError("unexpected type: {}".format(other))

        return self.system_name == other.system_name and self.user_key == other.user_key


# truncate and warn all strings in dictionary,
# ensure not shortening uuids
def strip_truncate_and_warn(d, root, length):
    for k, v in list(d.items()):
        if isinstance(v, dict):
            strip_truncate_and_warn(v, root, length)
        elif isinstance(v, str):
            v = d[k] = v.strip()
            if len(v) > length:
                v = d[k] = v[:length]
                logger.warning(
                    "truncating to %d key '%s' for" " uuid '%s' to value '%s'",
                    length,
                    k,
                    root["Uuid"],
                    v,
                )


@lru_cache
def os2mo_get(url, **params):
    #format url like {BASE}/service
    mora_base = get_os2sync_settings().mora_base
    
    url = url.format(BASE=f"{mora_base}/service")
    try:
        session = get_mo_session()
        r = session.get(url, params=params)
        r.raise_for_status()
        return r
    except Exception:
        logger.exception(url)
        raise


def has_kle():
    os2mo_config = os2mo_get("{BASE}/configuration").json()
    return os2mo_config["show_kle"]



def addresses_to_user(user, addresses, phone_scope_classes, email_scope_classes):
    # TODO: This looks like bucketing (more_itertools.bucket)
    emails, phones = [], []
    for address in addresses:
        if address["address_type"]["scope"] == "EMAIL":
            emails.append(address)
        if address["address_type"]["scope"] == "PHONE":
            phones.append(address)

    # find phone using prioritized/empty list of address_type uuids
    phone = choose_public_address(phones, phone_scope_classes)
    if phone:
        user["PhoneNumber"] = phone["name"]

    # find email using prioritized/empty list of address_type uuids
    email = choose_public_address(emails,email_scope_classes)
    if email:
        user["Email"] = email["name"]


def engagements_to_user(user, engagements, allowed_unitids):
    for e in sorted(engagements, key=lambda e: e["job_function"]["name"] + e["uuid"]):
        if e["org_unit"]["uuid"] in allowed_unitids:
            user["Positions"].append(
                {
                    "OrgUnitUuid": e["org_unit"]["uuid"],
                    "Name": e["job_function"]["name"],
                    #Only used to find primary engamgements work-address
                    "is_primary": e["is_primary"],
                }
            )


def try_get_ad_user_key(uuid: str) -> Optional[str]:
    """
    fetches all it-systems related to a user and return the ad-user_key if exists
    """
    it_response = os2mo_get("{BASE}/e/" + uuid + "/details/it").json()
    it_systems = IT.from_mo_json(it_response)
    ad_systems = list(filter(lambda x: x.system_name == AD_it_system, it_systems))

    # if no ad OR multiple
    if len(ad_systems) != 1:
        return
    return ad_systems[0].user_key

def get_work_address(positions, work_address_names):
    #find the primary engagement and lookup the addresses for that unit
    primary = filter(lambda e: e["is_primary"], positions)
    try:
        primary_eng = one(primary)
    except ValueError:
        logger.error("Could not get unique primary engagement, using first found position")
        primary_eng = first(positions)

    org_addresses = os2mo_get(
        "{BASE}/ou/" + primary_eng["OrgUnitUuid"] + "/details/address"
    ).json()
    #filter and sort based on settings and use the first match if any
    work_address = filter(lambda addr: addr["address_type"]["name"] in work_address_names , org_addresses)
    work_address = sorted(work_address, key=lambda a: work_address_names.index(a["address_type"]["name"]))
    chosen_work_address= first(work_address, default={})
    return chosen_work_address.get("name")

def get_sts_user(uuid, allowed_unitids, settings = None):
    employee = os2mo_get("{BASE}/e/" + uuid + "/").json()
    
    user = User(
        dict(
            uuid=uuid,
            candidate_user_id=try_get_ad_user_key(uuid),
            person=Person(employee, settings=settings),
        ),
        settings=settings,
    )

    sts_user = user.to_json()

    addresses_to_user(
        sts_user, os2mo_get("{BASE}/e/" + uuid + "/details/address").json(), phone_scope_classes=settings.os2sync_phone_scope_classes, email_scope_classes=settings.os2sync_email_scope_classes
    )
    #use calculate_primary flag to get the is_primary boolean used in getting work-address
    engagements = os2mo_get(
            "{BASE}/e/" + uuid + "/details/engagement?calculate_primary=true"
        ).json()
    engagements_to_user(
        sts_user,
        engagements,
        allowed_unitids,
    )

    # Optionally find the work address of employees primary engagement.
    work_address_names = settings.os2sync_employee_engagement_address
    if sts_user["Positions"] and work_address_names:
        sts_user["Location"] = get_work_address(sts_user["Positions"], work_address_names)
    truncate_length = max(36, settings.os2sync_truncate_length)
    strip_truncate_and_warn(sts_user, sts_user, length=truncate_length)
    return sts_user


def org_unit_uuids(**kwargs):
    org_uuid = one(os2mo_get("{BASE}/o/").json())["uuid"]
    return [
        ou["uuid"]
        for ou in os2mo_get(f"{{BASE}}/o/{org_uuid}/ou/", limit=999999, **kwargs).json()[
            "items"
        ]
    ]


def manager_to_orgunit(unit_uuid: UUID) -> UUID:
    manager = os2mo_get("{BASE}/ou/" + str(unit_uuid) + "/details/manager").json()
    if manager:
        return UUID(one(manager)["person"]["uuid"])


def itsystems_to_orgunit(orgunit, itsystems):
    for i in itsystems:
        orgunit["ItSystemUuids"].append(i["itsystem"]["uuid"])


def address_type_is(address, user_key=None, scope="TEXT"):
    return (
        address["address_type"]["user_key"] == user_key
        and address["address_type"]["scope"] == scope
    )


def addresses_to_orgunit(orgunit, addresses):
    for a in addresses:
        if a["address_type"]["scope"] == "EMAIL":
            orgunit["Email"] = a["name"]
        elif a["address_type"]["scope"] == "EAN":
            orgunit["Ean"] = a["name"]
        elif a["address_type"]["scope"] == "PHONE":
            orgunit["PhoneNumber"] = a["name"]
        elif a["address_type"]["scope"] == "DAR":
            orgunit["Post"] = a["name"]
        elif a["address_type"]["scope"] == "PNUMBER":
            orgunit["Location"] = a["name"]
        elif address_type_is(a, user_key="ContactOpenHours"):
            orgunit["ContactOpenHours"] = a["name"]
        elif address_type_is(a, user_key="DtrId"):
            orgunit["DtrId"] = a["name"]


def filter_kle(aspect: str, kle) -> List[UUID]:
    """Filters kle by aspect name

    KLE aspects can be "Udførende", "Ansvarlig" or "Indsigt"

    Returns:
        list of uuids
    """
    tasks_kle = filter(lambda k: one(k["kle_aspect"])["name"] == aspect, kle)
    task_uuids = set(k["kle_number"]["uuid"] for k in tasks_kle)
    return list(sorted(task_uuids))


def partition_kle(kle, use_contact_for_tasks) -> (List[UUID], List[UUID]):
    """Collect kle uuids according to kle_aspect.

    Default is to return all KLE uuids as Tasks,
    If the setting 'os2sync_use_contact_for_tasks' is set KLEs will be divided:

    * Aspect "Udførende" goes into "Tasks"
    * Aspect "Ansvarlig" goes into "ContactForTasks"

    Args:
        kle: A list of KLEs.

    Returns:
        Tuple(List, List)
    """

    if use_contact_for_tasks:
        tasks = filter_kle("Udførende", kle)
        ContactForTasks = filter_kle("Ansvarlig", kle)

        return tasks, ContactForTasks

    tasks = set()

    for k in kle:
        uuid = k["kle_number"]["uuid"]
        tasks.add(uuid)

    return list(sorted(tasks)), []


def kle_to_orgunit(org_unit: Dict, kle: Dict, use_contact_for_tasks) -> Dict:
    """Mutates the dict "org_unit" to include KLE data"""
    tasks, contactfortasks = partition_kle(kle, use_contact_for_tasks=use_contact_for_tasks)
    if tasks:
        org_unit["Tasks"] = tasks
    if contactfortasks:
        org_unit["ContactForTasks"] = contactfortasks


def is_ignored(unit, settings):
    """Determine if unit should be left out of transfer

    Args:
        unit: The organization unit to enrich with kle information.
        settings: a dictionary

    Returns:
        Boolean
    """

    return (
        unit.get("org_unit_level")
        and UUID(unit["org_unit_level"]["uuid"]) in settings.os2sync_ignored_unit_levels
    ) or (
        unit.get("org_unit_type")
        and UUID(unit["org_unit_type"]["uuid"]) in settings.os2sync_ignored_unit_types
    )


def get_sts_orgunit(uuid, settings):
    base = parent = os2mo_get("{BASE}/ou/" + uuid + "/").json()

    if is_ignored(base, settings):
        logger.info("Ignoring %r", base)
        return None
    top_unit_uuid = str(settings.os2sync_top_unit_uuid)
    if not parent["uuid"] == top_unit_uuid:
        while parent.get("parent"):
            if parent["uuid"] == top_unit_uuid:
                break
            parent = parent["parent"]

    if not parent["uuid"] == top_unit_uuid:
        # not part of right tree
        logger.debug(f"ignoring unit {uuid=}, as it is not a unit bellow {top_unit_uuid=}")
        return None

    sts_org_unit = {"ItSystemUuids": [], "Name": base["name"], "Uuid": uuid}

    if base.get("parent") and "uuid" in base["parent"]:
        sts_org_unit["ParentOrgUnitUuid"] = base["parent"]["uuid"]

    itsystems_to_orgunit(
        sts_org_unit, os2mo_get("{BASE}/ou/" + uuid + "/details/it").json()
    )
    addresses_to_orgunit(
        sts_org_unit,
        os2mo_get("{BASE}/ou/" + uuid + "/details/address").json(),
    )

    if settings.os2sync_sync_managers:
        manager_uuid = manager_to_orgunit(uuid)
        if manager_uuid:
            sts_org_unit["managerUuid"] = str(manager_uuid)

    # this is set by __main__
    if has_kle():
        kle_to_orgunit(
            sts_org_unit,
            os2mo_get("{BASE}/ou/" + uuid + "/details/kle").json(),
            use_contact_for_tasks=settings.os2sync_use_contact_for_tasks
        )

    # show_all_details(uuid,"ou")
    strip_truncate_and_warn(sts_org_unit, sts_org_unit, settings.os2sync_truncate_length)
    return sts_org_unit


def show_all_details(uuid, objtyp):
    import pprint

    print(" ---- details ----\n")
    for d, has_detail in (
        os2mo_get("{BASE}/" + objtyp + "/" + uuid + "/details").json().items()
    ):
        if has_detail:
            print("------------ detail ---- " + d)
            pprint.pprint(
                os2mo_get("{BASE}/" + objtyp + "/" + uuid + "/details/" + d).json()
            )
    print(" ---- end of details ----\n")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['SD_Lon',
 'SD_Lon.tests',
 'integrations',
 'integrations.SD_Lon',
 'integrations.SD_Lon.tests',
 'integrations.aarhus',
 'integrations.aarhus.tests',
 'integrations.ad_integration',
 'integrations.ad_integration.tests',
 'integrations.ad_integration.username_rules',
 'integrations.ballerup',
 'integrations.brøndby',
 'integrations.brønderslev',
 'integrations.calculate_primary',
 'integrations.calculate_primary.tests',
 'integrations.dar_helper',
 'integrations.egedal',
 'integrations.furesø',
 'integrations.gir.initial_read',
 'integrations.hjørring',
 'integrations.holstebro',
 'integrations.kle',
 'integrations.næstved',
 'integrations.opus',
 'integrations.opus.org_tree_print',
 'integrations.opus.tests',
 'integrations.os2sync',
 'integrations.os2sync.old_tests',
 'integrations.os2sync.tests',
 'integrations.rebild',
 'integrations.rundb',
 'integrations.rundb.tests',
 'integrations.sorø',
 'integrations.viborg']

package_data = \
{'': ['*'],
 'SD_Lon.tests': ['fixtures/tar_gz1/opt/dipex/os2mo-data-import-and-export/*',
                  'fixtures/tar_gz2/opt/dipex/os2mo-data-import-and-export/*'],
 'integrations': ['requirements/*'],
 'integrations.SD_Lon.tests': ['fixtures/tar_gz1/opt/dipex/os2mo-data-import-and-export/*',
                               'fixtures/tar_gz2/opt/dipex/os2mo-data-import-and-export/*'],
 'integrations.calculate_primary': ['requirements/*']}

install_requires = \
['Babel>=2.6.0,<3.0.0',
 'Jinja2>=2.10,<3.0',
 'MarkupSafe>=1.1.0,<2.0.0',
 'PyYAML>=6.0,<7.0',
 'Pygments>=2.3.1,<3.0.0',
 'SQLAlchemy>=1.4.31,<2.0.0',
 'Sphinx>=1.8.4,<2.0.0',
 'XlsxWriter>=3.0.2,<4.0.0',
 'aiohttp>=3.8.1,<4.0.0',
 'alabaster>=0.7.12,<0.8.0',
 'alchemy-mock>=0.4.3,<0.5.0',
 'alembic>=1.7.5,<2.0.0',
 'anytree>=2.6.0,<3.0.0',
 'asciitree>=0.3.3,<0.4.0',
 'black>=22.3.0,<23.0.0',
 'certifi>=2021.10.8,<2022.0.0',
 'chardet>=4.0.0,<5.0.0',
 'click-option-group>=0.5.3,<0.6.0',
 'click>=8.0.3,<9.0.0',
 'deepdiff>=5.7.0,<6.0.0',
 'docutils>=0.14,<0.15',
 'fastapi>=0.75.0,<0.76.0',
 'freezegun>=0.3.11,<0.4.0',
 'fs.smbfs>=1.0.3,<2.0.0',
 'google-cloud-storage>=2.1.0,<3.0.0',
 'httpx>=0.22.0,<0.23.0',
 'hypothesis>=6.36.1,<7.0.0',
 'idna>=3.3,<4.0',
 'imagesize>=1.1.0,<2.0.0',
 'isort==5.6.4',
 'jmespath>=0.10.0,<0.11.0',
 'jsonschema>=4.4.0,<5.0.0',
 'lora-utils>=0.1.0,<0.2.0',
 'lxml>=4.7.1,<5.0.0',
 'mimesis>=5.3.0,<6.0.0',
 'more-itertools>=8.12.0,<9.0.0',
 'mysqlclient>=2.1.0,<3.0.0',
 'os2mo-dar-client>=1,<2',
 'os2mo-data-import>=2.22.1,<3.0.0',
 'packaging>=19.0,<20.0',
 'pandas>=1.4.0,<2.0.0',
 'parameterized==0.7.4',
 'pika>=1.2.0,<2.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'pydantic>=1.8.1click,<2.0.0',
 'pymssql>=2.2.4,<3.0.0',
 'pyodbc>=4.0.32,<5.0.0',
 'pyparsing>=2.3.1,<3.0.0',
 'pytest-aioresponses>=0.2.0,<0.3.0',
 'pytest>=6.2.5,<7.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'pytz>=2021.3,<2022.0',
 'pywinrm[kerberos]>=0.4.2,<0.5.0',
 'ra-utils>=1.0.0,<2.0.0',
 'raclients>=1.2.4,<2.0.0',
 'ramodels==5.12.0',
 'recommonmark>=0.7.1,<0.8.0',
 'requests-kerberos>=0.14.0,<0.15.0',
 'requests>=2.21.0,<3.0.0',
 'retrying>=1.3.3,<2.0.0',
 'six>=1.12.0,<2.0.0',
 'snowballstemmer>=1.2.1,<2.0.0',
 'sphinx-rtd-theme>=1.0.0,<2.0.0',
 'sphinxcontrib-websupport>=1.1.0,<2.0.0',
 'spsftp>=0.0.1,<0.0.2',
 'tenacity>=8.0.1,<9.0.0',
 'tqdm>=4.62.0,<5.0.0',
 'types-python-dateutil>=2.8.9,<3.0.0',
 'wheel>=0.37.1,<0.38.0',
 'xlrd>=2.0.1,<3.0.0',
 'xmltodict>=0.12.0,<0.13.0']

setup_kwargs = {
    'name': 'dipex',
    'version': '2.22.2',
    'description': 'OS2mo-data-import-and-export',
    'long_description': '#################\nOS2MO Data Import\n#################\n\nMagentas officielle repo til integrationer og eksportfunktioner til OS2MO.\n\nFor spørgsmål til koden eller brug af den, er man velkommen til at kontakte\nMagenta ApS <info@magenta.dk>\n\nUsage\n-----\nStart en OS2mo stak vha. `docker-compose`, se detaljer her:\n* https://os2mo.readthedocs.io/en/1.16.1/dev/environment/docker.html?#docker-compose\n\nNår dette er sket, kan DIPEX udviklingsmiljøet startes med:\n```\ndocker-compose up -d --build\n```\nNår kommandoen er kørt færdig, kan man hoppe ind i containeren med:\n```\ndocker-compose exec dipex /bin/bash\n```\nDette giver en terminal i containeren, hvorfra diverse programmer kan køres.\nEt fælles entrypoint til programmerne findes ved at køre:\n```\npython3 metacli.py\n```\nForbindelsen imod OS2mo, kan testes med programmet: `check_connectivity`:\n```\npython3 metacli.py check_connectivity --mora-base http://mo\n```\n',
    'author': 'Magenta ApS',
    'author_email': 'info@magenta.dk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://magenta.dk/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from unittest.mock import patch
from uuid import uuid4

from ..ad_fix_enddate import CompareEndDate
from ..ad_fix_enddate import UpdateEndDate

# setup test variables and settings:
enddate_field = "enddate_field"
uuid_field = "uuid_field"
testcpr_field = "cpr_field"
testuuid = str(uuid4())
test_enddate = "2020-01-01"
test_search_base = "search_base"
test_settings = {
    "global": {"mora.base": "moramock"},
    "primary": {
        "cpr_field": testcpr_field,
        "search_base": test_search_base,
        "system_user": "username",
        "password": "password",
    },
}


@patch(
    "integrations.ad_integration.ad_common.AD._run_ps_script",
    return_value=[{enddate_field: "9999-12-31", uuid_field: testuuid, "CN": "TestCN"}],
)
@patch(
    "os2mo_helpers.mora_helpers.MoraHelper.read_user_engagement",
    return_value=[
        {"validity": {"from": "2019-01-01", "to": "2020-01-01"}, "is_primary": True}
    ],
)
@patch("integrations.ad_integration.ad_common.AD._create_session")
def test_ad_enddate_fixer(
    mock_ps_script,
    mock_read_user_engagements,
    mock_session,
):
    """Tests that user with no enddate in AD, but enddate in MO
    returns a ps-command to update AD.
    """
    c = CompareEndDate(enddate_field, uuid_field, test_settings)
    users = c.compare_mo()
    read_ps = f"""
        $User = "username"
        $PWord = ConvertTo-SecureString –String "password" –AsPlainText -Force
        $TypeName = "System.Management.Automation.PSCredential"
        $UserCredential = New-Object –TypeName $TypeName –ArgumentList $User, $PWord
        Get-ADUser -Filter \'{enddate_field} -like "9999-*"\' -SearchBase "{test_search_base}"  -Credential $usercredential -Properties cn,{uuid_field},{enddate_field} | ConvertTo-Json"""
    c._run_ps_script.assert_called_with(read_ps)
    assert list(users.keys()) == [testuuid]
    assert users[testuuid]["cn"] == "TestCN"

    u = UpdateEndDate(enddate_field, uuid_field, c.cpr_field, settings=test_settings)

    for uuid, end_date in u.get_changes(users):
        cmd = u.get_update_cmd(uuid, end_date)
        assert (
            cmd
            == f"""
        Get-ADUser  -SearchBase "{test_search_base}"  -Credential $usercredential -Filter \'{uuid_field} -eq "{testuuid}"\' |
        Set-ADUser  -Credential $usercredential -Replace @{{{enddate_field}="{test_enddate}"}} |
        ConvertTo-Json
        """
        )



from unittest import TestCase

import pytest

from ..ad_writer import EngagementDatesError
from ..ad_writer import LoraCacheSource
from ..ad_writer import MORESTSource
from ..utils import AttrDict
from .mocks import MockMORESTSource


class TestLoraCacheSource(TestCase):
    def setUp(self):
        self.user = self.setup_user()
        self.lc = AttrDict(
            {
                "users": {
                    self.user["uuid"]: [self.user],
                },
                "engagements": {
                    "key-1": [
                        # Current registration
                        {
                            "user": self.user["uuid"],
                            "from_date": "2020-01-01",
                            "to_date": None,
                        },
                        # Previous registration
                        {
                            "user": self.user["uuid"],
                            "from_date": "2019-01-01",
                            "to_date": "2021-01-01",
                        },
                    ],
                },
            }
        )
        self.lc_historic = AttrDict(
            {
                "engagements": {
                    # Add some historic engagements that must not be read by
                    # `get_engagement_dates`.
                    "key-1": [
                        # Current historic registration
                        {
                            "user": self.user["uuid"],
                            "from_date": "2010-01-01",
                            "to_date": "2019-12-31",
                        },
                    ]
                }
            }
        )

    def setup_user(self):
        return {
            "uuid": "some_uuid_here",
            "navn": "some_name some_lastname",
            "efternavn": "some_lastname",
            "fornavn": "some_name",
            "kaldenavn": "some_alias some_lastname_alias",
            "kaldenavn_fornavn": "some_alias",
            "kaldenavn_efternavn": "some_lastname_alias",
            "cpr": "some_cpr",
        }

    def test_read_user(self):
        datasource = self._get_datasource(None, None)
        result = datasource.read_user("some_uuid_here")
        self.assertEqual(
            result,
            {
                "uuid": "some_uuid_here",
                "name": "some_name some_lastname",
                "surname": "some_lastname",
                "givenname": "some_name",
                "nickname": "some_alias some_lastname_alias",
                "nickname_givenname": "some_alias",
                "nickname_surname": "some_lastname_alias",
                "cpr_no": "some_cpr",
            },
        )

    def test_get_engagement_dates(self):
        datasource = self._get_datasource("2020-01-01", None)
        result = datasource.get_engagement_dates(self.user["uuid"])
        self.assertEqual(
            [list(elem) for elem in result],  # consume each iterable in result
            [["2020-01-01"], [None]],
        )

    def test_get_engagement_endpoint_dates(self):
        datasource = self._get_datasource("2020-01-01", None)
        result = datasource.get_engagement_endpoint_dates(self.user["uuid"])
        # "to_date" of None must be converted into "9999-12-31"
        self.assertEqual(result, ("2020-01-01", "9999-12-31"))

    def _get_datasource(self, from_date, to_date):
        return LoraCacheSource(
            self.lc, self.lc_historic, MockMORESTSource(from_date, to_date)
        )


class TestEngagementDates:
    MOSource = MORESTSource(settings={"global": {"mora.base": "http://lol.money:1337"}})

    def test_valid_interval(self, monkeypatch):
        start, end = "2021-01-01", "2021-02-02"

        def _valid_dates(*args):
            return [start], [end]

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _valid_dates)
        assert self.MOSource.get_engagement_endpoint_dates("test") == (start, end)

    def test_sentinels(self, monkeypatch):
        start, end = "1930-01-01", "9999-12-31"

        def _falsy_dates(*args):
            return [None], [""]

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _falsy_dates)
        assert self.MOSource.get_engagement_endpoint_dates("test") == (start, end)

    def test_invalid_dates(self, monkeypatch):
        start, end = "2021-12-31", "2021-01-01"

        def _invalid_dates(*args):
            return [start], [end]

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _invalid_dates)
        with pytest.raises(EngagementDatesError, match=f"{start}.*{end}"):
            self.MOSource.get_engagement_endpoint_dates("test")

        def _empty_dates(*args):
            return [], []

        monkeypatch.setattr(self.MOSource, "get_engagement_dates", _empty_dates)
        with pytest.raises(EngagementDatesError, match="9999-12-31.*1930-01-01"):
            self.MOSource.get_engagement_endpoint_dates("test")



import logging
from uuid import UUID

import click
from pydantic import ValidationError

from integrations.SD_Lon.config import get_changed_at_settings
from integrations.SD_Lon.sd_common import sd_lookup


LOG_LEVEL = logging.DEBUG
LOG_FILE = "test_sd_connectivity.log"


def setup_logging():
    detail_logging = "sdCommon"
    for name in logging.root.manager.loggerDict:
        if name in detail_logging:
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        filename=LOG_FILE,
    )


class TestSdConnectivity(object):
    def __init__(self):
        self.validation_error = None
        try:
            self.settings = get_changed_at_settings()
        except ValidationError as err:
            self.validation_error = err

    def _check_sd_settings(self):
        print("Check settings...")

        if self.validation_error is None:
            print("SD settings OK")
        else:
            print("The following SD settings errors were detected:")
            print(self.validation_error)

        print("Done checking SD settings\n\n")

    def _check_contact_to_sd(self):
        print("Tjekker at vi har kontakt til SD:")
        params = {
            "UUIDIndicator": "true",
            "InstitutionIdentifier": self.settings.sd_institution_identifier,
        }
        try:
            institution_info = sd_lookup(
                "GetInstitution20111201",
                settings=self.settings,
                params=params,
                use_cache=False,
            )
        except Exception as e:
            print("Fejl i kontakt til SD Løn: {}".format(e))
            exit(1)

        try:
            institution = institution_info["Region"]["Institution"]
            institution_uuid = institution["InstitutionUUIDIdentifier"]
            UUID(institution_uuid, version=4)
            print(" * Korrekt kontakt til SD Løn")
        except Exception as e:
            msg = " * Fik forbindelse, men modtog ikke-korrekt svar fra SD: {}, {}"
            print(msg.format(institution_uuid, e))
            exit(1)

    def sd_check(self):
        self._check_sd_settings()
        self._check_contact_to_sd()


@click.command()
def check_connectivity():
    """Check SD configuration and connectivity."""
    tsc = TestSdConnectivity()
    tsc.sd_check()


if __name__ == "__main__":
    setup_logging()
    check_connectivity()



#
# Copyright (c) 2018, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import collections
import datetime
import json
import logging
import pathlib
from functools import partial
from operator import itemgetter

from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_setting
from tqdm import tqdm

from integrations.os2sync import config
from integrations.os2sync import lcdb_os2mo
from integrations.os2sync import os2mo
from integrations.os2sync import os2sync


logger = None  # set in main()
helper = None


def log_mox_config(settings):
    """It is imperative for log-forensics to have as
    much configuration as possible logged at program start
    and end.
    """

    logger.warning("-----------------------------------------")
    logger.warning("program configuration:")
    for k, v in settings:
        logger.warning("    %s=%r", k, v)


def log_mox_counters(counter):
    logger.info("-----------------------------------------")
    logger.info("program counters:")
    for k, v in sorted(counter.items()):
        logger.info("    %s: %r", k, v)


def sync_os2sync_orgunits(settings, counter, prev_date):
    logger.info("sync_os2sync_orgunits starting")

    logger.info(
        "sync_os2sync_orgunits getting " "all current organisational units from os2mo"
    )
    os2mo_uuids_present = set(os2mo.org_unit_uuids())

    logger.info(
        "sync_os2sync_orgunits getting " "units from os2mo from previous xfer date"
    )
    os2mo_uuids_past = set(os2mo.org_unit_uuids(at=prev_date))

    counter["Aktive Orgenheder fundet i OS2MO"] = len(os2mo_uuids_present)
    counter["Orgenheder tidligere"] = len(os2mo_uuids_past)

    logger.info(
        "sync_os2sync_orgunits deleting organisational "
        "units from os2sync if deleted in os2mo"
    )
    if len(os2mo_uuids_present):
        terminated_org_units = set(os2mo_uuids_past - os2mo_uuids_present)
        for uuid in tqdm(
            terminated_org_units, desc="Deleting terminated org_units", unit="org_unit"
        ):
            counter["Orgenheder som slettes i OS2Sync"] += 1
            os2sync.delete_orgunit(uuid)

    logger.info("sync_os2sync_orgunits upserting " "organisational units in os2sync")

    allowed_unitids = []
    for i in tqdm(os2mo_uuids_present, desc="Updating org_units", unit="org_unit"):
        sts_orgunit = os2mo.get_sts_orgunit(i, settings=settings)
        if sts_orgunit:
            allowed_unitids.append(i)
            counter["Orgenheder som opdateres i OS2Sync"] += 1
            os2sync.upsert_orgunit(sts_orgunit)
        elif settings.os2sync_autowash:
            counter["Orgenheder som slettes i OS2Sync"] += 1
            os2sync.delete_orgunit(i)

    logger.info("sync_os2sync_orgunits done")

    return set(allowed_unitids)


def sync_os2sync_users(settings, allowed_unitids, counter, prev_date):

    logger.info("sync_os2sync_users starting")

    logger.info(
        "sync_os2sync_users getting " "users from os2mo from previous xfer date"
    )
    os2mo_uuids_past = helper.read_all_users(at=prev_date)
    os2mo_uuids_past = set(map(itemgetter("uuid"), os2mo_uuids_past))

    logger.info("sync_os2sync_users getting list of users from os2mo")
    os2mo_uuids_present = helper.read_all_users()
    os2mo_uuids_present = set(map(itemgetter("uuid"), os2mo_uuids_present))

    counter["Medarbejdere fundet i OS2Mo"] = len(os2mo_uuids_present)
    counter["Medarbejdere tidligere"] = len(os2mo_uuids_past)

    logger.info("sync_os2sync_users deleting " "os2mo-deleted users in os2sync")

    if len(os2mo_uuids_present):
        terminated_users = set(os2mo_uuids_past - os2mo_uuids_present)
        for uuid in tqdm(
            terminated_users, desc="Deleting terminated users", unit="user"
        ):
            counter["Medarbejdere slettes i OS2Sync (del)"] += 1
            os2sync.delete_user(uuid)

    # insert/overwrite all users from os2mo
    # maybe delete if user has no more positions
    logger.info("sync_os2sync_users upserting os2sync users")

    for i in tqdm(os2mo_uuids_present, "Updating users", unit="user"):
        # medarbejdere er allerede omfattet af autowash
        # fordi de ikke får nogen 'Positions' hvis de ikke
        # har en ansættelse i en af allowed_unitids
        sts_user = os2mo.get_sts_user(i, allowed_unitids, settings=settings)

        if not sts_user["Positions"]:
            counter["Medarbejdere slettes i OS2Sync (pos)"] += 1
            os2sync.delete_user(i)
            continue

        os2sync.upsert_user(sts_user)
        counter["Medarbejdere overført til OS2SYNC"] += 1

    logger.info("sync_os2sync_users done")


def main(settings):
    # set warning-level for all loggers
    global logger
    [
        logging.getLogger(name).setLevel(logging.WARNING)
        for name in logging.root.manager.loggerDict
        if name != config.loggername
    ]

    logging.basicConfig(
        format=config.logformat,
        level=settings.os2sync_log_level,
        filename=settings.os2sync_log_file
    )
    logger = logging.getLogger(config.loggername)
    logger.setLevel(settings.os2sync_log_level)

    if settings.os2sync_use_lc_db:
        engine = lcdb_os2mo.get_engine()
        session = lcdb_os2mo.get_session(engine)
        os2mo.get_sts_user = partial(lcdb_os2mo.get_sts_user, session)
        os2mo.get_sts_orgunit = partial(lcdb_os2mo.get_sts_orgunit, session)

    prev_date = datetime.datetime.now() - datetime.timedelta(days=1)
    hash_cache_file = pathlib.Path(settings.os2sync_hash_cache)

    if hash_cache_file.exists():
        prev_date = datetime.datetime.fromtimestamp(hash_cache_file.stat().st_mtime)
    prev_date = prev_date.strftime("%Y-%m-%d")

    counter = collections.Counter()
    logger.info("mox_os2sync starting")
    log_mox_config(settings)

    if hash_cache_file and hash_cache_file.exists():
        os2sync.hash_cache.update(json.loads(hash_cache_file.read_text()))
    

    orgunit_uuids = sync_os2sync_orgunits(settings, counter, prev_date)
    sync_os2sync_users(settings, orgunit_uuids, counter, prev_date)

    if hash_cache_file:
        hash_cache_file.write_text(json.dumps(os2sync.hash_cache, indent=4))

    log_mox_counters(counter)
    log_mox_config(settings)
    logger.info("mox_os2sync done")


if __name__ == "__main__":
    settings = config.get_os2sync_settings()
    helper = MoraHelper(settings.mora_base)
    main(settings)



#!/usr/bin/env python3
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import datetime
import logging
import uuid
from operator import itemgetter
from typing import Any
from typing import Dict

import click
from anytree import Node
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_setting

from integrations import dawa_helper
from integrations.ad_integration import ad_reader
from integrations.SD_Lon.config import get_importer_settings
from integrations.SD_Lon.config import ImporterSettings
from integrations.SD_Lon.date_utils import format_date
from integrations.SD_Lon.date_utils import get_employment_dates
from integrations.SD_Lon.models import JobFunction
from integrations.SD_Lon.sd_common import calc_employment_id
from integrations.SD_Lon.sd_common import EmploymentStatus
from integrations.SD_Lon.sd_common import ensure_list
from integrations.SD_Lon.sd_common import generate_uuid
from integrations.SD_Lon.sd_common import read_employment_at
from integrations.SD_Lon.sd_common import sd_lookup
from integrations.SD_Lon.sd_common import skip_fictional_users
from os2mo_data_import import ImportHelper

LOG_LEVEL = logging.DEBUG
LOG_FILE = "mo_initial_import.log"

logger = logging.getLogger("sdImport")


class SdImport:
    # XXX: This does really expensive calls against SD in __init__, caution!

    def __init__(
        self,
        importer,
        settings: ImporterSettings,
        ad_info=None,
        org_only=False,
        org_id_prefix=None,
        employee_mapping=None,
    ):
        employee_mapping = employee_mapping or {}

        self.settings = settings

        self.base_url = "https://service.sd.dk/sdws/"
        self.address_errors: Dict[str, Dict] = {}

        self.importer = importer

        self.org_name = self.settings.municipality_name

        self.importer.add_organisation(
            identifier=self.org_name,
            user_key=self.org_name,
            municipality_code=self.settings.municipality_code,
        )

        self.org_id_prefix = org_id_prefix

        self.import_date = self.settings.sd_global_from_date.strftime("%d.%m.%Y")
        # List of job_functions that should be ignored.
        self.skip_job_functions = self.settings.sd_skip_employment_types

        # Whether to create SD-medarbejder associations
        self.create_associations = self.settings.sd_importer_create_associations

        # Whether to import email addresses for organisations
        self.create_email_addresses = self.settings.sd_importer_create_email_addresses

        # Whether to use <Employment><EmploymentDate> as engagement start date instead
        # of <Employment><EmploymentStatus><ActivationDate>.
        self.employment_date_as_engagement_start_date = (
            self.settings.sd_importer_employment_date_as_engagement_start_date
        )

        # CPR indexed dictionary of AD users
        self.ad_people: Dict[str, Dict] = {}
        self.employee_forced_uuids = employee_mapping
        self.ad_reader = None
        if isinstance(ad_info, ad_reader.ADParameterReader):
            self.ad_reader = ad_info
            self.importer.new_itsystem(identifier="AD", system_name="Active Directory")
            self.ad_reader.cache_all()

        self.nodes: Dict[str, Dict] = {}  # Will be populated when org-tree is created

        self.org_only = org_only
        if not org_only:
            self.add_people()

        self.info = self._read_department_info()

        self._add_classes()

    def _add_classes(self):
        # Format is facet -> list of entries
        # Entries are tuples on the format '(klasse_id, klasse, scope)'
        classes = {
            "manager_type": [
                ("leder_type", "Leder"),
            ],
            "org_unit_type": [
                ("Enhed", "Enhed"),
            ],
            "org_unit_address_type": [
                ("Pnummer", "Pnummer", "PNUMBER"),
                ("AddressMailUnit", "Postadresse", "DAR"),
                ("AdresseReturUnit", "Returadresse", "DAR"),
                ("AdresseHenvendelseUnit", "Henvendelsessted", "DAR"),
                ("PhoneUnit", "Telefon", "PHONE"),
                ("EmailUnit", "Email", "EMAIL"),
            ],
            "employee_address_type": [
                ("AdressePostEmployee", "Postadresse", "DAR"),
                ("PhoneEmployee", "Telefon", "PHONE"),
                ("MobilePhoneEmployee", "Mobiltelefon", "PHONE"),
                ("LocationEmployee", "Lokation", "TEXT"),
                ("EmailEmployee", "Email", "EMAIL"),
            ],
            "leave_type": [
                ("Orlov", "Orlov"),
            ],
            "engagement_type": [
                ("månedsløn", "Medarbejder (månedsløn)"),
                ("timeløn", "Medarbejder (timeløn)"),
                ("historisk", "Historisk ansættelse"),
            ],
            "primary_type": [
                ("Ansat", "Ansat", "3000"),
                ("status0", "Ansat - Ikke i løn", "1000"),
                ("non-primary", "Ikke-primær ansættelse", "0"),
                ("explicitly-primary", "Manuelt primær ansættelse", "5000"),
            ],
            "association_type": [
                ("SD-medarbejder", "SD-medarbejder"),
            ],
            "visibility": [
                ("Ekstern", "Må vises eksternt", "PUBLIC"),
                ("Intern", "Må vises internt", "INTERNAL"),
                ("Hemmelig", "Hemmelig", "SECRET"),
            ],
            "manager_level": [
                # XXX: Why 1040, 1035 and 1030?
                ("manager_1040", "Leder"),
                ("manager_1035", "Chef"),
                ("manager_1030", "Direktør"),
            ],
            "responsibility": [
                ("Lederansvar", "Lederansvar"),
            ],
        }

        for facet, klasses in classes.items():
            for klass in klasses:
                if len(klass) == 3:
                    klasse_id, klasse, scope = klass
                else:
                    klasse_id, klasse = klass
                    scope = "TEXT"
                self._add_klasse(klasse_id, klasse, facet, scope)

    def _update_ad_map(self, cpr):
        logger.debug("_update_ad_map called")
        self.ad_people[cpr] = {}
        if self.ad_reader:
            response = self.ad_reader.read_user(cpr=cpr, cache_only=True)
            if response:
                self.ad_people[cpr] = response

    def _read_department_info(self):
        """Load all department details and store for later user."""
        department_info = {}

        params = {
            "ActivationDate": self.import_date,
            "DeactivationDate": self.import_date,
            "ContactInformationIndicator": "true",
            "DepartmentNameIndicator": "true",
            "PostalAddressIndicator": "true",
            "ProductionUnitIndicator": "true",
            "UUIDIndicator": "true",
            "EmploymentDepartmentIndicator": "false",
        }
        departments = sd_lookup(
            "GetDepartment20111201", settings=self.settings, params=params
        )

        for department in departments["Department"]:
            uuid = department["DepartmentUUIDIdentifier"]
            if self.org_id_prefix:
                uuid = generate_uuid(uuid, self.org_id_prefix, self.org_name)

            department_info[uuid] = department
            unit_level = department["DepartmentLevelIdentifier"]
            if not self.importer.check_if_exists("klasse", unit_level):
                self._add_klasse(unit_level, unit_level, "org_unit_level", scope="TEXT")
        return department_info

    def _add_klasse(self, klasse_id, klasse, facet, scope="TEXT"):
        if isinstance(klasse_id, str):
            klasse_id = klasse_id.replace("&", "_")
        if not self.importer.check_if_exists("klasse", klasse_id):
            klasse_uuid = generate_uuid(klasse_id, self.org_id_prefix, self.org_name)
            self.importer.add_klasse(
                identifier=klasse_id,
                uuid=klasse_uuid,
                facet_type_ref=facet,
                user_key=str(klasse_id),
                scope=scope,
                title=klasse,
            )
        return klasse_id

    def _check_subtree(self, department, sub_tree):
        """Check if a department is member of given sub-tree"""
        while "DepartmentReference" in department:
            department = department["DepartmentReference"]
            dep_uuid = department["DepartmentUUIDIdentifier"]
            if self.org_id_prefix:
                dep_uuid = generate_uuid(dep_uuid, self.org_id_prefix)
            if dep_uuid == sub_tree:
                return True
        return False

    def _add_sd_department(
        self, department, contains_subunits=False, sub_tree=None, super_unit=None
    ):
        """
        Add add a deparment to MO. If the unit has parents, these will
        also be added
        :param department: The SD-department, including parent units.
        :param contains_subunits: True if the unit has sub-units.
        """
        ou_level = department["DepartmentLevelIdentifier"]
        if not self.org_id_prefix:
            unit_id = department["DepartmentUUIDIdentifier"]
            user_key = department["DepartmentIdentifier"]
        else:
            unit_id = generate_uuid(
                department["DepartmentUUIDIdentifier"], self.org_id_prefix
            )
            user_key = self.org_id_prefix + "_" + department["DepartmentIdentifier"]

        # parent_uuid = None
        parent_uuid = super_unit

        # If contain_subunits is true, this sub tree is a valid member
        import_unit = contains_subunits
        if "DepartmentReference" in department:
            if self._check_subtree(department, sub_tree):
                import_unit = True

            parent_uuid = department["DepartmentReference"]["DepartmentUUIDIdentifier"]
            if self.org_id_prefix:
                parent_uuid = self._generate_uuid(parent_uuid, self.org_id_prefix)
        else:
            import_unit = unit_id == sub_tree

        if not import_unit and sub_tree is not None:
            return

        info = self.info[unit_id]
        assert info["DepartmentLevelIdentifier"] == ou_level
        logger.debug("Add unit: {}".format(unit_id))
        if (
            (not contains_subunits)
            and (parent_uuid is super_unit)
            and self.importer.check_if_exists("organisation_unit", "OrphanUnits")
        ):
            parent_uuid = "OrphanUnits"

        date_from = info["ActivationDate"]
        # No units have termination dates: date_to is None
        if self.importer.check_if_exists("organisation_unit", unit_id):
            return
        else:
            self.importer.add_organisation_unit(
                identifier=unit_id,
                name=info["DepartmentName"],
                user_key=user_key,
                org_unit_level_ref=ou_level,
                type_ref="Enhed",
                date_from=date_from,
                uuid=unit_id,
                date_to=None,
                parent_ref=parent_uuid,
            )

        def import_email_addresses(info: Dict[str, Any]) -> None:
            if "ContactInformation" not in info:
                return
            if "EmailAddressIdentifier" not in info["ContactInformation"]:
                return

            emails = info["ContactInformation"]["EmailAddressIdentifier"]
            # filter empty entities
            emails = filter(lambda email: email.find("Empty") == -1, emails)
            for email in emails:
                self.importer.add_address_type(
                    organisation_unit=unit_id,
                    type_ref="EmailUnit",
                    value=email,
                    date_from=date_from,
                )

        def import_pnumber(info: Dict[str, Any]) -> None:
            if "ProductionUnitIdentifier" not in info:
                return
            self.importer.add_address_type(
                organisation_unit=unit_id,
                type_ref="Pnummer",
                value=info["ProductionUnitIdentifier"],
                date_from=date_from,
            )

        def import_postal_address(info: Dict[str, Any]) -> None:
            if "PostalAddress" not in info:
                return
            postal_address = info["PostalAddress"]

            if "StandardAddressIdentifier" not in postal_address:
                return
            address_string = postal_address["StandardAddressIdentifier"]

            if "PostalCode" not in postal_address:
                return
            zip_code = postal_address["PostalCode"]

            logger.debug("Look in Dawa: {}".format(address_string))
            dar_uuid = dawa_helper.dawa_lookup(address_string, zip_code)
            logger.debug("DAR: {}".format(dar_uuid))

            if dar_uuid is None:
                logger.error("Unable to lookup address: {}".format(address_string))
                self.address_errors[unit_id] = info
                return

            self.importer.add_address_type(
                organisation_unit=unit_id,
                type_ref="AddressMailUnit",
                value=dar_uuid,
                date_from=date_from,
            )

        if self.create_email_addresses:
            import_email_addresses(info)
        import_pnumber(info)
        import_postal_address(info)

        # Recursively create parents (higher level OUs)
        # Note: These do not have their own entry in SD
        if "DepartmentReference" in department:
            self._add_sd_department(
                department["DepartmentReference"],
                contains_subunits=True,
                sub_tree=sub_tree,
                super_unit=super_unit,
            )

    def _create_org_tree_structure(self):
        nodes = {}
        all_ous = self.importer.export("organisation_unit")
        new_ous = []
        for key, ou in all_ous.items():
            parent = ou.parent_ref
            if parent is None:
                uuid = key
                niveau = ou.org_unit_level_ref
                nodes[uuid] = Node(niveau, uuid=uuid)
            else:
                new_ous.append(ou)

        while len(new_ous) > 0:
            logger.info("Number of new ous: {}".format(len(new_ous)))
            print("Number of new ous: {}".format(len(new_ous)))
            all_ous = new_ous
            new_ous = []
            for ou in all_ous:
                parent = ou.parent_ref
                if parent in nodes.keys():
                    uuid = ou.uuid
                    niveau = ou.org_unit_level_ref
                    nodes[uuid] = Node(niveau, parent=nodes[parent], uuid=uuid)
                else:
                    new_ous.append(ou)
        return nodes

    def add_people(self):
        """Load all person details and store for later user"""
        params = {
            "StatusActiveIndicator": "true",
            "StatusPassiveIndicator": "false",
            "ContactInformationIndicator": "false",
            "PostalAddressIndicator": "false",
            "EffectiveDate": self.import_date,
        }
        active_people = sd_lookup(
            "GetPerson20111201", settings=self.settings, params=params
        )
        if not isinstance(active_people["Person"], list):
            active_people["Person"] = [active_people["Person"]]

        params["StatusActiveIndicator"] = False
        params["StatusPassiveIndicator"] = True
        passive_people = sd_lookup(
            "GetPerson20111201", settings=self.settings, params=params
        )
        if not isinstance(passive_people["Person"], list):
            passive_people["Person"] = [passive_people["Person"]]

        cprs = set(
            map(
                itemgetter("PersonCivilRegistrationIdentifier"), active_people["Person"]
            )
        )
        # Collect all people, prefering their active variants
        # TODO: Consider doing this with a dict keyed by cpr number
        people = active_people["Person"]
        for person in passive_people["Person"]:
            cpr = person["PersonCivilRegistrationIdentifier"]
            if cpr not in cprs:
                people.append(person)

        people = filter(skip_fictional_users, people)

        # TODO: Almost identitcal code exists in sd_changed_at's update_changed_persons
        for person in people:
            cpr = person["PersonCivilRegistrationIdentifier"]
            self._update_ad_map(cpr)

            given_name = person.get("PersonGivenName", "")
            sur_name = person.get("PersonSurnameName", "")

            uuid = self.employee_forced_uuids.get(cpr)
            logger.info("Employee in force list: {} {}".format(cpr, uuid))
            if uuid is None and "ObjectGuid" in self.ad_people[cpr]:
                uuid = self.ad_people[cpr]["ObjectGuid"]

            # Name is placeholder for initals, do not know which field to extract
            if "Name" in self.ad_people[cpr]:
                user_key = self.ad_people[cpr]["Name"]
            else:
                user_key = "{} {}".format(given_name, sur_name)

            self.importer.add_employee(
                name=(given_name, sur_name),
                identifier=cpr,
                cpr_no=cpr,
                user_key=user_key,
                uuid=uuid,
            )

            if "SamAccountName" in self.ad_people[cpr]:
                self.importer.join_itsystem(
                    employee=cpr,
                    user_key=self.ad_people[cpr]["SamAccountName"],
                    itsystem_ref="AD",
                    date_from="1930-01-01",
                )

    def create_ou_tree(self, create_orphan_container, sub_tree=None, super_unit=None):
        """Read all department levels from SD."""
        # TODO: Currently we can only read a top sub-tree
        if create_orphan_container:
            self._add_klasse("Orphan", "Virtuel Enhed", "org_unit_type")
            self.importer.add_organisation_unit(
                identifier="OrphanUnits",
                uuid="11111111-0000-0000-0000-111111111111",
                name="Forældreløse enheder",
                user_key="OrphanUnits",
                type_ref="Orphan",
                date_from="1930-01-01",
                date_to=None,
                parent_ref=None,
            )
        params = {
            "ActivationDate": self.import_date,
            "DeactivationDate": self.import_date,
            "UUIDIndicator": "true",
        }
        organisation = sd_lookup(
            "GetOrganization20111201", settings=self.settings, params=params
        )
        departments = organisation["Organization"]["DepartmentReference"]

        for department in departments:
            self._add_sd_department(
                department, sub_tree=sub_tree, super_unit=super_unit
            )
        self.nodes = self._create_org_tree_structure()

    def create_employees(self):
        logger.info("Create employees")

        effective_date = datetime.datetime.strptime(self.import_date, "%d.%m.%Y").date()

        logger.info("Get active people from SD...")
        active_people = ensure_list(
            read_employment_at(
                effective_date,
                settings=self.settings,
                status_passive_indicator=False,
            )
        )

        logger.info("Get passive people from SD...")
        passive_people = ensure_list(
            read_employment_at(
                effective_date,
                settings=self.settings,
                status_active_indicator=False,
                status_passive_indicator=True,
            )
        )

        # TODO: condense the two calls above into one and then create the variables
        # active_people and passive_people by using the partition function from
        # more_itertools

        logger.info("Create employees from active SD employees...")
        self._create_employees(active_people)

        logger.info("Create employees from passive SD employees...")
        self._create_employees(passive_people, skip_manager=True)

    def _create_employees(self, persons, skip_manager=False):
        people = filter(skip_fictional_users, persons)
        for person in people:
            self.create_employee(person, skip_manager=skip_manager)

    def create_employee(self, person, skip_manager=False):
        logger.debug(79 * "-")
        logger.debug("Person object to create: {}".format(person))

        cpr = person["PersonCivilRegistrationIdentifier"]
        employments = ensure_list(person["Employment"])

        for employment in employments:
            status = EmploymentStatus(
                employment["EmploymentStatus"]["EmploymentStatusCode"]
            )

            # Job_position_id: Klassificeret liste over stillingstyper.
            # job_name: Fritekstfelt med stillingsbetegnelser.
            job_position_id = employment["Profession"]["JobPositionIdentifier"]
            if job_position_id in self.skip_job_functions:
                continue
            job_name = employment["Profession"].get("EmploymentName", job_position_id)

            occupation_rate = float(employment["WorkingTime"]["OccupationRate"])

            employment_id = calc_employment_id(employment)

            # TODO: Identital code to this exists in sd_changed_at
            split = self.settings.sd_monthly_hourly_divide
            if employment_id["value"] < split:
                engagement_type_ref = "månedsløn"
            elif (split - 1) < employment_id["value"] < 999999:
                engagement_type_ref = "timeløn"
            else:  # This happens if EmploymentID is not a number
                engagement_type_ref = "engagement_type" + job_position_id
                self._add_klasse(
                    engagement_type_ref, job_position_id, "engagement_type"
                )
                logger.info("Non-numeric id. Job pos id: {}".format(job_position_id))

            primary_type_ref = "non-primary"
            # If status 0, uncondtionally override
            if status == EmploymentStatus.AnsatUdenLoen:
                primary_type_ref = "status0"

            job_function_type = self.settings.sd_job_function
            if job_function_type == JobFunction.employment_name:
                job_func_ref = self._add_klasse(
                    job_name, job_name, "engagement_job_function"
                )
            elif job_function_type == JobFunction.job_position_identifier:
                job_func_ref = self._add_klasse(
                    job_position_id, job_position_id, "engagement_job_function"
                )

            emp_dep = employment["EmploymentDepartment"]
            unit = emp_dep["DepartmentUUIDIdentifier"]

            date_from, date_to = get_employment_dates(
                employment,
                self.settings.sd_importer_employment_date_as_engagement_start_date,
            )

            date_from_str = format_date(date_from)
            date_to_str = format_date(date_to)
            if date_to == datetime.datetime(9999, 12, 31, 0, 0):
                date_to_str = None

            sd_employment_id = employment_id["id"]
            logger.info(
                f"Validty for {sd_employment_id}: from: {date_from_str}, "
                f"to: {date_to_str}"
            )

            if date_to <= date_from:
                logger.warning(f"Skip creating employment for id: {sd_employment_id}")
                continue

            original_unit = unit
            # Remove this to remove any sign of the employee from the
            # lowest levels of the org
            if self.create_associations:
                self.importer.add_association(
                    employee=cpr,
                    user_key=employment_id["id"],
                    organisation_unit=original_unit,
                    association_type_ref="SD-medarbejder",
                    date_from=date_from_str,
                    date_to=date_to_str,
                )

            # Employees are not allowed to be in these units (allthough
            # we do make an association). We must instead find the lowest
            # higher level to put she or he.
            too_deep = self.settings.sd_import_too_deep
            while self.nodes[unit].name in too_deep:
                unit = self.nodes[unit].parent.uuid

            ext_field = self.settings.sd_employment_field
            extention = {}
            if ext_field is not None:
                extention[ext_field] = job_name

            # Generate UUID for engagement here since we need this UUID when/if
            # creating a leave below
            engagement_uuid = str(uuid.uuid4())

            self.importer.add_engagement(
                employee=cpr,
                uuid=engagement_uuid,
                user_key=employment_id["id"],
                organisation_unit=unit,
                job_function_ref=job_func_ref,
                fraction=int(occupation_rate * 1000000),
                primary_ref=primary_type_ref,
                engagement_type_ref=engagement_type_ref,
                date_from=date_from_str,
                date_to=date_to_str,
                **extention,
            )
            if status == EmploymentStatus.Orlov:
                self.importer.add_leave(
                    employee=cpr,
                    engagement_uuid=engagement_uuid,
                    leave_type_ref="Orlov",
                    date_from=employment["EmploymentStatus"]["ActivationDate"],
                    date_to=date_to_str,
                )

            # These job functions will normally (but necessarily)
            #  correlate to a manager position
            if job_position_id in ["1040", "1035", "1030"]:
                self.importer.add_manager(
                    employee=cpr,
                    organisation_unit=unit,
                    manager_level_ref="manager_" + job_position_id,
                    address_uuid=None,  # Manager address is not used
                    manager_type_ref="leder_type",
                    responsibility_list=["Lederansvar"],
                    date_from=date_from_str,
                    date_to=date_to_str,
                )


@click.group()
def cli():
    for name in logging.root.manager.loggerDict:
        if name in (
            "sdImport",
            "sdCommon",
            "AdReader",
            "moImporterMoraTypes",
            "moImporterMoxTypes",
            "moImporterUtilities",
            "moImporterHelpers",
        ):
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        filename=LOG_FILE,
    )


@cli.command()
@click.option(
    "--org-only",
    is_flag=True,
    default=False,
    type=click.BOOL,
    help="Only import organisation structure",
)
@click.option(
    "--mora-base",
    default=load_setting("mora.base", "http://localhost:5000"),
    help="URL for OS2mo.",
)
@click.option(
    "--mox-base",
    default=load_setting("mox.base", "http://localhost:8080"),
    help="URL for LoRa.",
)
def full_import(org_only: bool, mora_base: str, mox_base: str):
    """Tool to do an initial full import."""
    # Check connection to MO before we fire requests against SD
    mh = MoraHelper(mora_base)
    if not mh.check_connection():
        raise click.ClickException("No MO reply, aborting.")

    importer = ImportHelper(
        create_defaults=True,
        mox_base=mox_base,
        mora_base=mora_base,
        store_integration_data=False,
        seperate_names=True,
    )
    sd = SdImport(
        importer, settings=get_importer_settings(), org_only=org_only, ad_info=None
    )

    sd.create_ou_tree(create_orphan_container=False, sub_tree=None, super_unit=None)
    if not org_only:
        sd.create_employees()

    importer.import_all()
    print("IMPORT DONE")


if __name__ == "__main__":
    cli()



import unittest
from unittest import mock

from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st
from parameterized import parameterized

from ..ad_exceptions import ImproperlyConfigured
from ..user_names import UserNameGen
from ..user_names import UserNameGenMethod2
from ..user_names import UserNameGenPermutation
from ..user_names import UserNameSet
from ..user_names import UserNameSetCSVFile
from ..user_names import UserNameSetInAD
from .mocks import MockADParameterReader
from .name_simulator import create_name


class TestUserNameGen(unittest.TestCase):
    _load_setting_path = "integrations.ad_integration.user_names.load_setting"

    @parameterized.expand(
        [
            ("UserNameGenMethod2", UserNameGenMethod2),
            ("UserNameGenUnknown", UserNameGenMethod2),
            ("UserNameGenPermutation", UserNameGenPermutation),
        ]
    )
    def test_override_implementation(self, name, expected_class):
        """Test that `get_implementation` returns the expected implementation
        configured by setting 'integrations.ad_writer.user_names.class'.
        """
        with mock.patch(self._load_setting_path, return_value=lambda: name):
            impl = UserNameGen.get_implementation()
            self.assertIsInstance(impl, expected_class)
            self.assertEqual(len(impl._loaded_occupied_name_sets), 0)

    def test_default_if_no_settings(self):
        """Test that `get_implementation` returns the default implementation
        if 'settings.json' could not be read.
        """

        def raise_filenotfound():
            raise FileNotFoundError()

        with mock.patch(self._load_setting_path, return_value=raise_filenotfound):
            impl = UserNameGen.get_implementation()
            self.assertIsInstance(impl, UserNameGenMethod2)
            self.assertEqual(len(impl._loaded_occupied_name_sets), 0)

    def test_load_occupied_names(self):
        """Test that we can configure `UserNameGen` to load one or more
        `UserNameSet` classes and add their individual sets of occupied names.
        """
        # Get `UserNameGen` implementation which loads occupied usernames from
        # `UserNameSet` (= an empty implementation.)
        impl = self._get_instance("UserNameSet")

        # Assert that we loaded two username sets (from AD and from the
        # `UserNameSet` specified in settings.)
        self.assertEqual(len(impl._loaded_occupied_name_sets), 2)

        # Assert that the first username set is the AD username set.
        self.assertIsInstance(impl._loaded_occupied_name_sets[0], UserNameSetInAD)

        # Assert that the second username set is the one we specified
        # via settings, and is an empty set.
        self.assertIsInstance(impl._loaded_occupied_name_sets[1], UserNameSet)
        self.assertSetEqual(set(impl._loaded_occupied_name_sets[1]), set())

        # Assert that the total set of occupied usernames is equal to
        # the usernames "found" by our mock `ADParameterReader`.
        self.assertSetEqual(
            impl.occupied_names,
            set(impl._loaded_occupied_name_sets[0]),
        )

    def test_load_occupied_names_invalid_name_raises(self):
        with self.assertRaises(ImproperlyConfigured):
            self._get_instance("InvalidUserNameSetName")

    def _get_instance(self, usernameset_class_name: str) -> "UserNameGen":
        cls_names = [usernameset_class_name]
        settings = {
            f"{UserNameGen._setting_prefix}.extra_occupied_name_classes": cls_names
        }
        with self._patch_settings(settings):
            # Use mock `ADParameterReader` in `UserNameSetInAD`
            with mock.patch(
                "integrations.ad_integration.user_names.ADParameterReader",
                new=MockADParameterReader,
            ):
                impl = UserNameGen.get_implementation()
                impl.load_occupied_names()
                return impl

    def _patch_settings(self, settings: dict):
        return mock.patch(
            self._load_setting_path,
            new=lambda name, default=None: lambda: settings.get(name, default),
        )


# PIMJE is missing in specification documen
pmj_first = [
    "pmunj",
    "pmuje",
    "pjens",
    "pimuj",
    "pimje",
    "pijen",
    "piamj",
    "piaje",
    "piamu",
    "pimun",
    "pmunk",
    "jense",
    "mjens",
    "mujen",
    "munje",
    "munkj",
]
pmj_second = ["pmuj", "pimj", "pjen", "pije", "piaj", "pmje"]
pmj_third = ["pmj", "pje", "pij", "pim", "pmu"]
pmj_fourth = [
    "pmu2j",
    "pm2je",
    "p2jen",
    "p2muj",
    "p2mje",
    "pim2j",
    "pi2je",
    "pi2mj",
    "pia2j",
    "piam2",
    "pimu2",
    "pmun2",
]

pmj_fifth = [
    "pmj2",
    "pmu2",
    "pij2",
    "pim2",
    "pia2",
    "mun2",
    "jen2",
    "pm2j",
    "pi2j",
    "pi2m",
    "p2mj",
    "p2je",
    "p2mu",
]
pmj_sixth = [
    "pmunkj",
    "pimunj",
    "piamuj",
    "pmunje",
    "pmujen",
    "pmjens",
    "pjense",
    "pijens",
    "piajen",
]
pmj = [pmj_first, pmj_second, pmj_third, pmj_fourth, pmj_fifth, pmj_sixth]

akjpa_first = [
    "akjpa",
    "akjea",
    "akria",
    "akrja",
    "akrpa",
    "akpea",
    "ajepa",
    "ajpea",
    "apeta",
    "akran",
    "akjan",
    "akpan",
    "ajean",
    "ajpan",
    "apean",
    "aande",
    "ankpa",
    "ankja",
    "ankra",
    "anjpa",
    "anjea",
    "anpea",
    "ankan",
    "anjan",
    "anpan",
    "anand",
    "andka",
    "andja",
    "andpa",
    "andan",
    "andkr",
    "andkj",
    "andkp",
    "andje",
    "andjp",
    "andpe",
    "andea",
    "andek",
    "andej",
    "andep",
    "ankjp",
    "ankje",
    "ankrp",
    "ankrj",
    "ankri",
    "ankpe",
    "anjpe",
    "anjep",
    "anjen",
    "anpet",
    "akjpe",
    "akjep",
    "akrjp",
    "akrip",
    "akrij",
    "akrje",
    "akjen",
    "akrpe",
    "akpet",
    "ajpet",
    "ajepe",
    "ajenp",
    "ajens",
    "apete",
    "akris",
    "ander",
    "kande",
    "krand",
    "krian",
    "krisa",
    "kjand",
    "kjean",
    "kjena",
    "kjpan",
    "kjpea",
    "kpand",
    "kpean",
    "kpeta",
    "jande",
    "jeand",
    "jenan",
    "jensa",
    "jpand",
    "jpean",
    "jpeta",
    "pande",
    "peand",
    "petan",
    "petea",
]
akjpa_second = [
    "akra",
    "akja",
    "akpa",
    "ajea",
    "ajpa",
    "apea",
    "anka",
    "anja",
    "anpa",
    "aand",
    "anan",
    "anda",
    "akan",
    "ajan",
    "apan",
]
akjpa_third = [
    "aka",
    "aja",
    "apa",
    "aan",
    "ana",
    "ank",
    "anj",
    "anp",
    "akr",
    "akj",
    "akp",
    "aje",
    "ajp",
    "ape",
]
akjpa_fourth = [
    "akj2a",
    "akr2a",
    "akp2a",
    "aje2a",
    "ajp2a",
    "ape2a",
    "ak2an",
    "aj2an",
    "ap2an",
    "a2and",
    "a2kpa",
    "a2kja",
    "a2kra",
    "a2jpa",
    "a2jea",
    "a2pea",
    "a2kan",
    "a2jan",
    "a2pan",
    "ank2a",
    "anj2a",
    "anp2a",
    "an2an",
    "an2ka",
    "an2ja",
    "an2pa",
    "and2a",
    "ande2",
    "andk2",
    "andj2",
    "andp2",
    "ankr2",
    "ankj2",
    "ankp2",
    "anje2",
    "anjp2",
    "anpe2",
    "akjp2",
    "akri2",
    "akrj2",
    "akrp2",
    "akje2",
    "akpe2",
    "ajen2",
    "ajep2",
    "ajpe2",
    "apet2",
]
akjpa_fifth = [
    "aka2",
    "aja2",
    "apa2",
    "akj2",
    "akp2",
    "akr2",
    "ajp2",
    "aje2",
    "ape2",
    "ana2",
    "ank2",
    "anj2",
    "anp2",
    "and2",
    "kjp2",
    "kje2",
    "kri2",
    "jep2",
    "jpe2",
    "pet2",
    "ak2a",
    "aj2a",
    "ap2a",
    "an2a",
    "ak2j",
    "ak2p",
    "aj2p",
    "an2k",
    "an2j",
    "an2p",
    "a2ka",
    "a2ja",
    "a2pa",
    "a2an",
    "a2kj",
    "a2kp",
    "a2kr",
    "a2jp",
    "a2je",
    "a2pe",
]
akjpa_sixth = [
    "akrisa",
    "akrija",
    "akripa",
    "akrjea",
    "akrjpa",
    "akrpea",
    "akjena",
    "akjepa",
    "akjpea",
    "ajensa",
    "ajenpa",
    "ajepea",
    "ajpeta",
    "apetea",
    "ankria",
    "ankrja",
    "ankrpa",
    "ankjea",
    "ankjpa",
    "anjena",
    "anjepa",
    "anjpea",
    "anpeta",
    "andkra",
    "andkja",
    "andkpa",
    "andjea",
    "andjpa",
    "andpea",
    "andeka",
    "andeja",
    "andepa",
    "akrian",
    "akrjan",
    "akrpan",
    "akjean",
    "akjpan",
    "akpean",
    "ajenan",
    "ajepan",
    "ajpean",
    "apetan",
    "akrand",
    "akjand",
    "akpand",
    "ajeand",
    "ajpand",
    "apeand",
    "akande",
    "ajande",
    "apande",
    "aander",
    "anande",
    "andand",
    "andean",
    "andera",
]
akjpa = [
    akjpa_first,
    akjpa_second,
    akjpa_third,
    akjpa_fourth,
    akjpa_fifth,
    akjpa_sixth,
]

kj_first = ["kjens", "kajen", "karje", "karij", "jense"]
kj_second = ["kjen", "kaje", "karj"]
kj_third = ["kje", "kaj"]
kj_fourth = ["k2jen", "ka2je", "kar2j", "kari2"]
kj_fifth = ["kaj2", "kar2", "jen2", "ka2j", "k2je"]
kj_sixth = ["kjense", "kajens", "karjen", "karije", "karinj"]
kj = [kj_first, kj_second, kj_third, kj_fourth, kj_fifth, kj_sixth]

kmj_first = [
    "kmunj",
    "kmuje",
    "kjens",
    "kamuj",
    "kamje",
    "kajen",
    "karmj",
    "karje",
    "karmu",
    "karij",
    "karim",
    "kamun",
    "kmunk",
    "jense",
    "mjens",
    "mujen",
    "munje",
    "munkj",
]
kmj_second = ["kmuj", "kamj", "kjen", "kaje", "karj", "kmje"]
kmj_third = ["kmj", "kje", "kaj", "kam", "kmu"]
kmj_fourth = [
    "kmu2j",
    "km2je",
    "k2jen",
    "k2muj",
    "k2mje",
    "kam2j",
    "ka2je",
    "ka2mj",
    "kar2j",
    "kari2",
    "karm2",
    "kamu2",
    "kmun2",
]
kmj_fifth = [
    "kmj2",
    "kmu2",
    "kaj2",
    "kam2",
    "kar2",
    "mun2",
    "jen2",
    "km2j",
    "ka2j",
    "ka2m",
    "k2mj",
    "k2je",
    "k2mu",
]
kmj_sixth = [
    "kmunkj",
    "kamunj",
    "karmuj",
    "karimj",
    "kmunje",
    "kmujen",
    "kmjens",
    "kjense",
    "kajens",
    "karjen",
    "karije",
    "karinj",
]
kmj = [kmj_first, kmj_second, kmj_third, kmj_fourth, kmj_fifth, kmj_sixth]

ooha_first = [
    "oohoa",
    "oosta",
    "oosha",
    "oosar",
    "oohar",
    "ohoar",
    "oaroe",
    "oloha",
    "olosa",
    "olhoa",
    "oloar",
    "olhar",
    "olaro",
    "oleoa",
    "oleha",
    "olear",
    "oleos",
    "oleoh",
    "oleho",
    "oloho",
    "olosh",
    "olost",
    "olhos",
    "oosth",
    "oosho",
    "oohos",
    "ohost",
    "ooste",
    "osaro",
    "ostar",
    "ostea",
    "oharo",
    "ohosa",
    "haroe",
    "hoaro",
    "hosar",
    "hosta",
]
ooha_second = [
    "oosa",
    "ooha",
    "ohoa",
    "oloa",
    "olha",
    "oaro",
    "olar",
    "olea",
    "ooar",
    "ohar",
]
ooha_third = ["ooa", "oha", "oar", "ola", "olo", "olh", "oos", "ooh", "oho"]
ooha_fourth = [
    "ooh2a",
    "oos2a",
    "oho2a",
    "oo2ar",
    "oh2ar",
    "o2aro",
    "o2oha",
    "o2osa",
    "o2hoa",
    "o2oar",
    "o2har",
    "olo2a",
    "olh2a",
    "ol2ar",
    "ol2oa",
    "ol2ha",
    "ole2a",
    "oleo2",
    "oleh2",
    "olos2",
    "oloh2",
    "olho2",
    "oost2",
    "oosh2",
    "ooho2",
    "ohos2",
]
ooha_fifth = [
    "ooa2",
    "oha2",
    "ooh2",
    "oos2",
    "oho2",
    "ola2",
    "olo2",
    "olh2",
    "ole2",
    "ost2",
    "aro2",
    "oo2a",
    "oh2a",
    "ol2a",
    "oo2h",
    "ol2o",
    "ol2h",
    "o2oa",
    "o2ha",
    "o2ar",
    "o2oh",
    "o2os",
    "o2ho",
]
ooha_sixth = [
    "oostea",
    "oostha",
    "ooshoa",
    "oohosa",
    "ohosta",
    "olosta",
    "olosha",
    "olohoa",
    "olhosa",
    "oleosa",
    "oleoha",
    "olehoa",
    "oostar",
    "ooshar",
    "oohoar",
    "ohosar",
    "oosaro",
    "ooharo",
    "ohoaro",
    "ooaroe",
    "oharoe",
    "olaroe",
    "olearo",
]
ooha = [
    ooha_first,
    ooha_second,
    ooha_third,
    ooha_fourth,
    ooha_fifth,
    ooha_sixth,
]


class TestUserNameGenMethod2(unittest.TestCase):
    def _test_person(self, name, reference, max_level):
        name_creator = UserNameGenMethod2()
        success = True
        for level in range(0, max_level):
            for correct_user_name in reference[level]:
                user_name = name_creator.create_username(name)
                if user_name != correct_user_name:
                    success = False
                    print("Got: {}, expected: {}".format(user_name, correct_user_name))
        return success

    def test_pia_munk_jensen(self):
        name = ["Pia", "Munk", "Jensen"]
        for i in range(1, 7):
            test_result = self._test_person(name, pmj, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_anders_kristian_jens_peter_andersen(self):
        name = ["Anders", "Kristian", "Jens", "Peter", "Andersen"]
        for i in range(1, 7):
            test_result = self._test_person(name, akjpa, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_karina_jensen(self):
        name = ["Karina", "Jensen"]
        for i in range(1, 7):
            test_result = self._test_person(name, kj, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_karina_munk_jensen(self):
        name = ["Karina", "Munk", "Jensen"]
        for i in range(1, 7):
            test_result = self._test_person(name, kmj, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_Olê_Østergård_Høst_Ærøe(self):
        name = ["Olê", "Østergård", "Høst", "Ærøe"]
        for i in range(1, 7):
            test_result = self._test_person(name, ooha, i)
            self.assertTrue(test_result, "Priority: {}".format(i))

    def test_too_many_karina_jensens(self):
        """
        Tests behaviour of creating so many identical names, that
        the algorithm runs out of options. For 'Karina Jensen' this
        should happen after 89 attempts.
        """
        name = ["Karina", "Jensen"]
        name_creator = UserNameGenMethod2()
        for i in range(1, 88):
            name_creator.create_username(name)
        with self.assertRaisesRegex(RuntimeError, "Failed to create user name"):
            name_creator.create_username(name)

    def test_dry_run(self):
        """
        Tests that with dry_run set to True, we can keep adding the
        same person and never run into an answer of None.
        """
        name = ["Karina", "Jensen"]
        name_creator = UserNameGenMethod2()
        for i in range(1, 100):
            user_name = name_creator.create_username(name, dry_run=True)
        self.assertFalse(user_name is None)

    def test_multiple_names(self):
        """
        Test that we can create a large number of random names without hitting a
        bug that stops the program. Test succeeds if the code does not raise an
        exception.
        """
        name_creator = UserNameGenMethod2()
        for i in range(0, 250):
            name = create_name()
            name_creator.create_username(name)

    def test_name_fixer(self):
        """
        Test that the name fixer allows a-z and does not allow values outside this
        range.
        """
        name_creator = UserNameGenMethod2()
        name = ["Anders", "abzæ-{øå", "Andersen"]
        fixed_name = name_creator._name_fixer(name)
        expected_name = ["Anders", "abzaoa", "Andersen"]
        self.assertTrue(fixed_name == expected_name)


CONSONANTS = st.characters(
    whitelist_characters="bcdfghjklmnpqrstvwxz",
    whitelist_categories=(),
)


class TestUserNameGenPermutation(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.instance = UserNameGenPermutation()

    @settings(max_examples=1000, deadline=None)
    @given(st.lists(st.text(min_size=1, alphabet=CONSONANTS), min_size=0))
    def test_valid_input(self, name):
        # If `name` has at least one item, each item being a string of at least
        # 0 consonants, we should be able to create a username.
        self.instance.create_username(name)

    def test_suffix_and_permutation_index_increments(self):
        name = ["B", "C", "D"]
        expected_permutations = ("bcd", "bdc", "cbd", "cdb", "dbc", "dcb")
        for expected_permutation in expected_permutations:
            for expected_suffix in range(1, 10):
                username = self.instance.create_username(name)
                self.assertEqual(
                    username,
                    "%s%d" % (expected_permutation, expected_suffix),
                )

    def test_skips_names_already_taken(self):
        name = ["Lille", "Claus", "Her"]
        self.instance.add_occupied_names({"lch1", "lch4"})
        for expected_username in ("lch2", "lch3", "lch5"):
            username = self.instance.create_username(name)
            self.assertEqual(username, expected_username)

    @parameterized.expand(
        [
            # name, expected feed
            (["Bac", "Cab"], ["B", "C", "c", "b"]),  # unpadded
            (["Baa", "Caa"], ["B", "C", "x"]),  # padded with one 'x'
        ]
    )
    def test_feed_is_padded(self, name, expected_feed):
        self.assertEqual(self.instance._get_feed(name), expected_feed)

    def test_raises_indexerror_when_feed_exhausted(self):
        # If we keep generating usernames from a very small feed of only three
        # consonants, we will run out of new usernames when we have generated
        # (3! * 9) = (6 * 9) = 54 usernames from that particular feed.
        for num in range(55):  # sequence from 0 to and including 54
            try:
                self.instance.create_username(["B", "C", "D"])
            except IndexError:
                self.assertEqual(num, 54)


class TestUserNameSetCSVFile(unittest.TestCase):
    csv_path = "some/fs/path"
    csv_lines = [
        "%s,foo" % UserNameSetCSVFile._column_name,
        "abcd1234,",
        "efgh5678",
    ]

    def test_can_read_csv(self):
        instance = self._get_instance()
        instance._mock_open.assert_called_once_with(
            self.csv_path,
            "r",
            encoding=UserNameSetCSVFile._encoding,
        )
        self.assertEqual(instance._usernames, {"abcd1234", "efgh5678"})

    def test_contains(self):
        instance = self._get_instance()
        self.assertIn("abcd1234", instance)
        self.assertNotIn("abc123", instance)

    def test_iter(self):
        instance = self._get_instance()
        self.assertSetEqual(instance._usernames, set(instance))

    def _get_instance(self):
        load_setting_path = "integrations.ad_integration.user_names.load_setting"
        with mock.patch(load_setting_path, return_value=lambda: self.csv_path):
            with mock.patch("io.open") as mock_open:
                mock_open.return_value.__enter__.return_value = self.csv_lines
                instance = UserNameSetCSVFile()
                instance._mock_open = mock_open
                return instance



from datetime import date
from datetime import datetime
from uuid import uuid4

import los_files
import los_leder
from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st

from .helpers import HelperMixin
from .strategies import csv_buf_from_model


class _TestableManagerImporter(los_leder.ManagerImporter):  # type: ignore
    """Defanged `ManagerImporter` with stub versions of `handle_{create,edit,terminate}`
    methods, as well as `cache_cpr`.
    """

    def __init__(self):
        super().__init__()
        self._handler_calls = []

    def cache_cpr(self):
        pass

    async def handle_create(self, filename: str, filedate: datetime):
        self._handler_calls.append(("handle_create", filename, filedate))

    async def handle_edit(self, filename: str, filedate: datetime):
        self._handler_calls.append(("handle_edit", filename, filedate))

    async def handle_terminate(self, filename: str, filedate: datetime):
        self._handler_calls.append(("handle_terminate", filename, filedate))


class TestParseManagerCreateCSV:
    @given(csv_buf_from_model(model=los_leder.ManagerCreate))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_leder.ManagerCreate)


class TestParseManagerEditCSV:
    @given(csv_buf_from_model(model=los_leder.ManagerEdit))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_leder.ManagerEdit)


class TestParseManagerTerminateCSV:
    @given(csv_buf_from_model(model=los_leder.ManagerTerminate))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_leder.ManagerTerminate)


class TestManagerImporter(HelperMixin):
    _employee_uuid = uuid4()

    @given(st.builds(los_leder.ManagerCreate), st.datetimes())
    def test_handle_create(
        self, instance: los_leder.ManagerCreate, filedate: datetime  # type: ignore
    ):
        importer, mock_csv, mock_session = self._setup(instance)
        with mock_csv, mock_session:
            with self._mock_create_details() as mock_create_details:
                self._run_until_complete(
                    importer.handle_create("unused_filename.csv", filedate)
                )
                self._assert_payload_contains(
                    mock_create_details,
                    {
                        "type": "manager",
                        "person": {"uuid": str(self._employee_uuid)},
                        "validity": {"from": date.today().isoformat(), "to": None},
                    },
                )

    @given(st.builds(los_leder.ManagerEdit), st.datetimes())
    def test_handle_edit_creates_new_detail(
        self, instance: los_leder.ManagerEdit, filedate: datetime  # type: ignore
    ):
        importer, mock_csv, mock_session = self._setup(instance)
        with mock_csv, mock_session:
            with self._mock_create_details() as mock_create_details:
                with self._mock_edit_details() as mock_edit_details:
                    with self._mock_lookup_organisationfunktion():
                        self._run_until_complete(
                            importer.handle_edit("unused_filename.csv", filedate)
                        )
                        self._assert_payload_contains(
                            mock_create_details,
                            {
                                "type": "manager",
                                "person": {"uuid": str(self._employee_uuid)},
                                "validity": {
                                    "from": date.today().isoformat(),
                                    "to": None,
                                },
                            },
                        )
                        self._assert_payload_empty(mock_edit_details)

    @given(st.builds(los_leder.ManagerEdit), st.datetimes())
    def test_handle_edit_updates_existing_detail(
        self, instance: los_leder.ManagerEdit, filedate: datetime  # type: ignore
    ):
        importer, mock_csv, mock_session = self._setup(instance)
        with mock_csv, mock_session:
            with self._mock_create_details() as mock_create_details:
                with self._mock_edit_details() as mock_edit_details:
                    orgfunk_uuid = importer._generate_rel_uuid(instance)
                    with self._mock_lookup_organisationfunktion(
                        return_value=set([str(orgfunk_uuid)])
                    ):
                        self._run_until_complete(
                            importer.handle_edit("unused_filename.csv", filedate)
                        )
                        self._assert_payload_empty(mock_create_details)
                        self._assert_payload_contains(
                            mock_edit_details,
                            {
                                "type": "manager",
                                "uuid": str(orgfunk_uuid),
                                "data": {
                                    "type": "manager",
                                    "uuid": str(orgfunk_uuid),
                                    "org_unit": {"uuid": str(instance.org_uuid)},
                                    "person": {"uuid": str(self._employee_uuid)},
                                    "validity": {
                                        "from": filedate.date().isoformat(),
                                    },
                                    "manager_level": {
                                        "uuid": str(instance.manager_level_uuid)
                                    },
                                    "manager_type": {
                                        "uuid": str(instance.manager_type_uuid)
                                    },
                                    "responsibility": [
                                        {"uuid": str(instance.responsibility_uuid)}
                                    ],
                                },
                            },
                        )

    @given(st.builds(los_leder.ManagerTerminate), st.datetimes())
    def test_handle_terminate(
        self, instance: los_leder.ManagerTerminate, filedate: datetime  # type: ignore
    ):
        importer, mock_csv, mock_session = self._setup(instance)
        with mock_csv, mock_session:
            with self._mock_terminate_details() as mock_terminate_details:
                self._run_until_complete(
                    importer.handle_terminate("unused_filename.csv", filedate)
                )
                self._assert_payload_contains(
                    mock_terminate_details,
                    {
                        "type": "manager",
                        "uuid": str(importer._generate_rel_uuid(instance)),
                        "validity": {"to": filedate.date().isoformat()},
                    },
                )

    @given(st.builds(los_leder.ManagerCreate))
    def test_generate_manager_payload_unmatched_cpr(self, instance):
        # Create importer with empty `cpr_cache`
        importer = los_leder.ManagerImporter()
        # Generate manager payload which will contain invalid person UUID due
        # to not finding the UUID in the empty `cpr_cache`.
        payload = importer.generate_manager_payload(instance)
        assert payload["person"]["uuid"] == "None"

    def test_cache_cpr(self):
        employee = {"cpr_no": "0101012222", "uuid": uuid4()}
        importer = los_leder.ManagerImporter()
        with self._mock_lookup_employees(return_value=[employee]):
            importer.cache_cpr()
            assert importer.cpr_cache == {employee["cpr_no"]: employee["uuid"]}

    def test_run(self):
        mock_last_import = datetime(2020, 1, 1)
        mock_filenames = [
            "Leder_nye_20200101_000001.csv",
            "Leder_ret_20200101_000001.csv",
            "Leder_luk_20200101_000001.csv",
        ]
        expected_filedate = datetime(2020, 1, 1, 0, 0, 1)
        with self._mock_get_import_filenames(
            mock_filenames
        ) as mock_get_import_filenames:
            importer = _TestableManagerImporter()
            # Run method under test
            self._run_until_complete(importer.run(mock_last_import))
            # Assert `get_import_filenames` was called (without arguments)
            mock_get_import_filenames.assert_called_once_with()
            # Assert that `handle_{create,edit,terminate}` were called with the
            # expected arguments.
            methods = ("handle_create", "handle_edit", "handle_terminate")
            for method, filename in zip(methods, mock_filenames):
                expected_call = (method, filename, expected_filedate)
                assert expected_call in importer._handler_calls

    def _setup(self, instance):
        importer = los_leder.ManagerImporter()
        importer.cpr_cache = {instance.cpr: str(self._employee_uuid)}
        mock_csv = self._mock_read_csv(instance)
        mock_session = self._mock_get_client_session()
        return importer, mock_csv, mock_session

    def _assert_payload_contains(self, call, expected):
        call.assert_called_once()
        # Second arg of first (and only) call contains the payloads
        payloads = list(call.call_args[0][1])
        assert len(payloads) == 1
        # Assert content and structure of first (and only) payload
        for key, value in expected.items():
            assert payloads[0][key] == value

    def _assert_payload_empty(self, call):
        call.assert_called_once()
        payloads = list(call.call_args[0][1])
        assert len(payloads) == 0


class TestManagerImporterCacheCpr(HelperMixin):
    @settings(max_examples=1000, deadline=None)
    @given(
        # Build simulated return value of `util.lookup_employees`
        st.lists(
            # Employees with both UUID and CPR
            st.fixed_dictionaries({"uuid": st.text(), "cpr_no": st.text()})
            |
            # Employees with only UUID
            st.fixed_dictionaries({"uuid": st.text()})
        )
    )
    def test_cache_cpr_handles_missing_cpr(self, mock_employee_list):
        instance = los_leder.ManagerImporter()
        with self._mock_lookup_employees(return_value=mock_employee_list):
            instance.cache_cpr()
            unique_mock_cprs = set(
                emp["cpr_no"] for emp in mock_employee_list if emp.get("cpr_no")
            )
            assert set(instance.cpr_cache.keys()) == unique_mock_cprs
